import os
from logging import getLogger
from typing import NamedTuple, Protocol, TypeVar, Sequence, runtime_checkable

import cv2
import numpy as np
from cv2.typing import MatLike, Size
from skimage.metrics import structural_similarity

from .core import Image, unify_image
from ..util import Rect, Point
from .debug import result as debug_result, debug, img
from .preprocessor import PreprocessorProtocol

logger = getLogger(__name__)

class TemplateNoMatchError(Exception):
    """模板未找到异常。"""
    def __init__(self, image: MatLike | Image, template: MatLike | str | Image):
        self.image = image
        self.template = template
        super().__init__(f"Template not found: {template}")

@runtime_checkable
class ResultProtocol(Protocol):
    @property
    def rect(self) -> Rect:
        """结果区域。左上角坐标和宽高。"""
        ...


class TemplateMatchResult(NamedTuple):
    score: float
    position: Point
    """结果位置。左上角坐标。"""
    size: Size
    """输入模板的大小。宽高。"""

    @property
    def rect(self) -> Rect:
        """结果区域。左上角坐标和宽高。"""
        return (self.position[0], self.position[1], self.size[0], self.size[1])
    
    @property
    def right_bottom(self) -> Point:
        """结果右下角坐标。"""
        return (self.position[0] + self.size[0], self.position[1] + self.size[1])

class MultipleTemplateMatchResult(NamedTuple):
    score: float
    position: Point
    """结果位置。左上角坐标。"""
    size: Size
    """命中模板的大小。宽高。"""
    index: int
    """命中模板在列表中的索引。"""

    @property
    def rect(self) -> Rect:
        """结果区域。左上角坐标和宽高。"""
        return (self.position[0], self.position[1], self.size[0], self.size[1])
    
    @property
    def right_bottom(self) -> Point:
        """结果右下角坐标。"""
        return (self.position[0] + self.size[0], self.position[1] + self.size[1])

    @classmethod
    def from_template_match_result(cls, result: TemplateMatchResult, index: int):
        return cls(
            score=result.score,
            position=result.position,
            size=result.size,
            index=index
        )

class CropResult(NamedTuple):
    score: float
    position: Point
    size: Size
    image: MatLike

    @property
    def rect(self) -> Rect:
        return (self.position[0], self.position[1], self.size[0], self.size[1])

def _draw_result(image: MatLike, matches: Sequence[ResultProtocol] | ResultProtocol | None) -> MatLike:
    """在图像上绘制匹配结果的矩形框。"""
    if matches is None:
        return image
    if isinstance(matches, ResultProtocol):
        matches = [matches]
    result_image = image.copy()
    for match in matches:
        cv2.rectangle(result_image, match.rect, (0, 0, 255), 2)
    return result_image

def _img2str(image: MatLike | str | Image | None) -> str:
    if image is None:
        return 'None'
    if isinstance(image, str):
        try:
            return os.path.relpath(image)
        except ValueError:
            # ValueError: path is on mount 'C:', start on mount 'E:'
            # 程序路径与资源路径不在同一个地方的情况
            return image
    elif isinstance(image, Image):
        return f'<Image: {image.name} at {image.path}>'
    else:
        return '<opencv Mat>'

def _imgs2str(images: Sequence[MatLike | str | Image | None] | None) -> str:
    if images is None:
        return 'None'
    return ', '.join([_img2str(image) for image in images])

def _result2str(result: TemplateMatchResult | MultipleTemplateMatchResult | None) -> str:
    if result is None:
        return 'None'
    return f'{result.rect} {result.score}'

def _results2str(results: Sequence[TemplateMatchResult | MultipleTemplateMatchResult] | None) -> str:
    if results is None:
        return 'None'
    return ', '.join([_result2str(result) for result in results])

# TODO: 应该把 template_match 和 find、wait、expect 等函数的公共参数提取出来
# TODO: 需要在调试结果中输出 preprocessors 处理后的图像
def template_match(
    template: MatLike | str | Image,
    image: MatLike | str | Image,
    mask: MatLike | str | Image | None = None,
    *,
    transparent: bool = False,
    threshold: float = 0.8,
    max_results: int = 5,
    remove_duplicate: bool = True,
    colored: bool = False,
    preprocessors: list[PreprocessorProtocol] | None = None,
) -> list[TemplateMatchResult]:
    """
    寻找模板在图像中的位置。

    .. note::
        `mask` 和 `transparent` 参数不能同时使用。
        如果使用透明图像，所有透明像素必须为 100% 透明，不能包含半透明像素。

    :param template: 模板图像，可以是图像路径或 cv2.Mat。
    :param image: 图像，可以是图像路径或 cv2.Mat。
    :param mask: 掩码图像，可以是图像路径或 cv2.Mat。
    :param transparent: 若为 True，则认为输入模板是透明的，并自动将透明模板转换为 Mask 图像。
    :param threshold: 阈值，默认为 0.8。
    :param max_results: 最大结果数，默认为 1。
    :param remove_duplicate: 是否移除重复结果，默认为 True。
    :param colored: 是否匹配颜色，默认为 False。
    :param preprocessors: 预处理列表，默认为 None。
    """
    # 统一参数
    template = unify_image(template, transparent)
    image = unify_image(image)
    if transparent is True and mask is not None:
        raise ValueError('mask and transparent cannot be used together')
    if mask is not None:
        mask = unify_image(mask)
        mask = cv2.threshold(mask, 127, 255, cv2.THRESH_BINARY)[1]
    if transparent is True:
        # https://stackoverflow.com/questions/57899997/how-to-create-mask-from-alpha-channel-in-opencv
        # 从透明图像中提取 alpha 通道作为 mask
        mask = cv2.threshold(template[:, :, 3], 0, 255, cv2.THRESH_BINARY)[1]
        template = template[:, :, :3]
    # 预处理
    if preprocessors is not None:
        for preprocessor in preprocessors:
            image = preprocessor.process(image)
            template = preprocessor.process(template)
            if mask is not None:
                mask = preprocessor.process(mask)
    # 匹配模板
    if mask is not None:
        # https://stackoverflow.com/questions/35642497/python-opencv-cv2-matchtemplate-with-transparency
        # 使用 Mask 时，必须使用 TM_CCORR_NORMED 方法
        result = cv2.matchTemplate(image, template, cv2.TM_CCORR_NORMED, mask=mask)
    else:
        result = cv2.matchTemplate(image, template, cv2.TM_CCOEFF_NORMED)
    
    # ========== 整理结果 ==========
    # 去重、排序、转换为 TemplateMatchResult

    # 获取所有大于阈值的匹配结果并按分数排序
    h, w = template.shape[:2]
    matches = []
    if remove_duplicate:
        # 创建一个掩码来标记已匹配区域
        used_mask = np.zeros((image.shape[0], image.shape[1]), np.uint8)
    
    # 获取所有匹配点并按分数从高到低排序
    match_points = np.where(result >= threshold)
    scores = result[match_points]
    sorted_indices = np.argsort(-scores)  # 降序排序
    
    for idx in sorted_indices:
        y, x = match_points[0][idx], match_points[1][idx]
        score = float(scores[idx])
        
        # 去重
        if remove_duplicate:
            # 获取匹配区域的中心点
            center_x = x + w // 2
            center_y = y + h // 2
            
            # 如果中心点已被标记，跳过此匹配
            if used_mask[center_y, center_x] == 255:
                continue
                
            # 标记整个匹配区域
            used_mask[y:y+h, x:x+w] = 255
        
        # 颜色匹配
        if colored:
            img1, img2 = image[y:y+h, x:x+w], template
            if mask is not None:
                # 如果用了 Mask，需要裁剪出 Mask 区域，其余部分置黑
                img1 = cv2.bitwise_and(img1, img1, mask=mask)
                img2 = cv2.bitwise_and(img2, img2, mask=mask)

            if not hist_match(img1, img2, (0, 0, w, h)):
                continue
        
        matches.append(TemplateMatchResult(
            score=score,
            position=(int(x), int(y)),
            size=(int(w), int(h))
        ))
        
        # 如果达到最大结果数，提前结束
        if max_results > 0 and len(matches) >= max_results:
            break
    
    return matches

def hist_match(
    image: MatLike | str,
    template: MatLike | str,
    rect: Rect | None = None,
    threshold: float = 0.9,
) -> bool:
    """
    对输入图像的矩形部分与模板进行颜色直方图匹配。
    将图像分为上中下三个区域，分别计算直方图并比较相似度。

    https://answers.opencv.org/question/59027/template-matching-using-color/

    :param image: 输入图像
    :param template: 模板图像
    :param rect: 输入图像中待匹配的矩形区域
    :param threshold: 相似度阈值，默认为 0.8
    :return: 是否匹配成功
    """
    # 统一参数
    image = unify_image(image)
    template = unify_image(template)

    # 从图像中裁剪出矩形区域
    if rect is None:
        roi = image
    else:
        x, y, w, h = rect
        roi = image[y:y+h, x:x+w]

    # 确保尺寸一致
    if roi.shape != template.shape:
        # roi = cv2.resize(roi, (template.shape[1], template.shape[0]))
        raise ValueError('Expected two images with the same size.')

    # 将图像分为上中下三个区域
    h = roi.shape[0]
    h_band = h // 3
    bands_roi = [
        roi[0:h_band],
        roi[h_band:2*h_band],
        roi[2*h_band:h]
    ]
    bands_template = [
        template[0:h_band],
        template[h_band:2*h_band],
        template[2*h_band:h]
    ]

    # 计算每个区域的直方图
    total_score = 0
    for roi_band, template_band in zip(bands_roi, bands_template):
        hist_roi = cv2.calcHist([roi_band], [0, 1, 2], None, [8, 8, 8], [0, 256, 0, 256, 0, 256])
        hist_template = cv2.calcHist([template_band], [0, 1, 2], None, [8, 8, 8], [0, 256, 0, 256, 0, 256])
        
        # 归一化直方图
        cv2.normalize(hist_roi, hist_roi)
        cv2.normalize(hist_template, hist_template)
        
        # 计算直方图相似度
        score = cv2.compareHist(hist_roi, hist_template, cv2.HISTCMP_CORREL)
        total_score += score

    # 计算平均相似度
    avg_score = total_score / 3
    return avg_score >= threshold

def find_all_crop(
    image: MatLike | str | Image,
    template: MatLike | str | Image,
    mask: MatLike | str | Image | None = None,
    transparent: bool = False,
    threshold: float = 0.8,
    *,
    colored: bool = False,
    remove_duplicate: bool = True,
    preprocessors: list[PreprocessorProtocol] | None = None,
) -> list[CropResult]:
    """
    指定一个模板，在输入图像中寻找其出现的所有位置，并裁剪出结果。

    :param image: 图像，可以是图像路径或 cv2.Mat。
    :param template: 模板图像，可以是图像路径或 cv2.Mat。
    :param mask: 掩码图像，可以是图像路径或 cv2.Mat。
    :param transparent: 若为 True，则认为输入模板是透明的，并自动将透明模板转换为 Mask 图像。
    :param threshold: 阈值，默认为 0.8。
    :param colored: 是否匹配颜色，默认为 False。
    :param remove_duplicate: 是否移除重复结果，默认为 True。
    :param preprocessors: 预处理列表，默认为 None。
    """
    matches = template_match(
        template,
        image,
        mask,
        transparent=transparent,
        threshold=threshold,
        max_results=-1,
        remove_duplicate=remove_duplicate,
        colored=colored,
        preprocessors=preprocessors,
    )
    # logger.debug(
    #     f'find_all_crop(): template: {_img2str(template)} image: {_img2str(image)} mask: {_img2str(mask)} '
    #     f'matches: {_results2str(matches)}'
    # )
    return [CropResult(
        match.score,
        match.position,
        match.size,
        image[match.rect[1]:match.rect[1]+match.rect[3], match.rect[0]:match.rect[0]+match.rect[2]] # type: ignore
    ) for match in matches]

def find(
    image: MatLike,
    template: MatLike | str | Image,
    mask: MatLike | str | Image | None = None,
    *,
    transparent: bool = False,
    threshold: float = 0.8,
    debug_output: bool = True,
    colored: bool = False,
    remove_duplicate: bool = True,
    preprocessors: list[PreprocessorProtocol] | None = None,
) -> TemplateMatchResult | None:
    """
    指定一个模板，在输入图像中寻找其出现的第一个位置。

    :param image: 图像，可以是图像路径或 cv2.Mat。
    :param template: 模板图像，可以是图像路径或 cv2.Mat。
    :param mask: 掩码图像，可以是图像路径或 cv2.Mat。
    :param transparent: 若为 True，则认为输入模板是透明的，并自动将透明模板转换为 Mask 图像。
    :param threshold: 阈值，默认为 0.8。
    :param debug_output: 是否输出调试信息，默认为 True。
    :param colored: 是否匹配颜色，默认为 False。
    :param remove_duplicate: 是否移除重复结果，默认为 True。
    :param preprocessors: 预处理列表，默认为 None。
    """
    matches = template_match(
        template,
        image,
        mask,
        transparent=transparent,
        threshold=threshold,
        max_results=1,
        remove_duplicate=remove_duplicate,
        colored=colored,
        preprocessors=preprocessors,
    )
    # logger.debug(
    #     f'find(): template: {_img2str(template)} image: {_img2str(image)} mask: {_img2str(mask)} '
    #     f'matches: {_results2str(matches)}'
    # )
    # 调试输出
    if debug.enabled and debug_output:
        result_image = _draw_result(image, matches)
        result_text = f"template: {img(template)} \n"
        result_text += f"matches: {len(matches)} \n"
        for match in matches:
            result_text += f"score: {match.score} position: {match.position} size: {match.size} \n"
        debug_result(
            'image.find',
            [result_image, image],
            result_text
        )
    return matches[0] if len(matches) > 0 else None

def find_all(
    image: MatLike,
    template: MatLike | str | Image,
    mask: MatLike | str | Image | None = None,
    *,
    transparent: bool = False,
    threshold: float = 0.8,
    remove_duplicate: bool = True,
    colored: bool = False,
    debug_output: bool = True,
    preprocessors: list[PreprocessorProtocol] | None = None,
) -> list[TemplateMatchResult]:
    """
    指定一个模板，在输入图像中寻找其出现的所有位置。

    :param image: 图像，可以是图像路径或 cv2.Mat。
    :param template: 模板图像，可以是图像路径或 cv2.Mat。
    :param mask: 掩码图像，可以是图像路径或 cv2.Mat。
    :param transparent: 若为 True，则认为输入模板是透明的，并自动将透明模板转换为 Mask 图像。
    :param threshold: 阈值，默认为 0.8。
    :param remove_duplicate: 是否移除重复结果，默认为 True。
    :param colored: 是否匹配颜色，默认为 False。
    :param preprocessors: 预处理列表，默认为 None。
    """
    results = template_match(
        template,
        image,
        mask,
        transparent=transparent,
        threshold=threshold,
        max_results=-1,
        remove_duplicate=remove_duplicate,
        colored=colored,
        preprocessors=preprocessors,
    )
    # logger.debug(
    #     f'find_all(): template: {_img2str(template)} image: {_img2str(image)} mask: {_img2str(mask)} '
    #     f'matches: {_results2str(results)}'
    # )
    if debug.enabled and debug_output:
        result_image = _draw_result(image, results)
        debug_result(
            'image.find_all',
            [result_image, image],
            f"template: {img(template)} \n"
            f"matches: {len(results)} \n"
        )
    return results

def find_multi(
    image: MatLike,
    templates: Sequence[MatLike | str | Image],
    masks: Sequence[MatLike | str | Image | None] | None = None,
    *,
    transparent: bool = False,
    threshold: float = 0.8,
    colored: bool = False,
    remove_duplicate: bool = True,
    preprocessors: list[PreprocessorProtocol] | None = None,
) -> MultipleTemplateMatchResult | None:
    """
    指定多个模板，在输入图像中逐个寻找模板，返回第一个匹配到的结果。

    :param image: 图像，可以是图像路径或 cv2.Mat。
    :param templates: 模板图像列表，可以是图像路径或 cv2.Mat。
    :param masks: 掩码图像列表，可以是图像路径或 cv2.Mat。
    :param transparent: 若为 True，则认为输入模板是透明的，并自动将透明模板转换为 Mask 图像。
    :param threshold: 阈值，默认为 0.8。
    :param colored: 是否匹配颜色，默认为 False。
    :param remove_duplicate: 是否移除重复结果，默认为 True。
    :param preprocessors: 预处理列表，默认为 None。
    """
    ret = None
    if masks is None:
        _masks = [None] * len(templates)
    else:
        _masks = masks
    for index, (template, mask) in enumerate(zip(templates, _masks)):
        find_result = find(
            image,
            template,
            mask,
            transparent=transparent,
            threshold=threshold,
            colored=colored,
            debug_output=False,
            remove_duplicate=remove_duplicate,
            preprocessors=preprocessors,
        )
        # 调试输出
        if find_result is not None:
            ret = MultipleTemplateMatchResult(
                score=find_result.score,
                position=find_result.position,
                size=find_result.size,
                index=index
            )
            break
    # logger.debug(
    #     f'find_multi(): templates: {_imgs2str(templates)} images: {_img2str(image)} masks: {_imgs2str(masks)} '
    #     f'result: {_result2str(ret)}'
    # )
    if debug.enabled:
        msg = (
            "<table class='result-table'>" +
            "<tr><th>Template</th><th>Mask</th><th>Result</th></tr>" +
            "\n".join([
                f"<tr><td>{img(t)}</td><td>{img(m)}</td><td>{'✓' if ret and t == templates[ret.index] else '✗'}</td></tr>"
                for i, (t, m) in enumerate(zip(templates, _masks))
            ]) +
            "</table>\n"
        )
        debug_result(
            'image.find_multi',
            [_draw_result(image, ret), image],
            msg
        )
    return ret

def find_all_multi(
    image: MatLike,
    templates: list[MatLike | str | Image],
    masks: list[MatLike | str | Image | None] | None = None,
    *,
    transparent: bool = False,
    threshold: float = 0.8,
    colored: bool = False,
    remove_duplicate: bool = True,
    preprocessors: list[PreprocessorProtocol] | None = None,
) -> list[MultipleTemplateMatchResult]:
    """
    指定多个模板，在输入图像中逐个寻找模板，返回所有匹配到的结果。

    此函数等价于
    ```python
    result = []
    for template in templates:
        result.append(find_all(template, ...))
    ```

    :param image: 图像，可以是图像路径或 cv2.Mat。
    :param templates: 模板图像列表，可以是图像路径或 cv2.Mat。
    :param masks: 掩码图像列表，可以是图像路径或 cv2.Mat。
    :param transparent: 若为 True，则认为输入模板是透明的，并自动将透明模板转换为 Mask 图像。
    :param threshold: 阈值，默认为 0.8。
    :param colored: 是否匹配颜色，默认为 False。
    :param remove_duplicate: 是否移除重复结果，默认为 True。
    :param preprocessors: 预处理列表，默认为 None。
    :return: 匹配到的一维结果列表。
    """
    ret: list[MultipleTemplateMatchResult] = []
    if masks is None:
        _masks = [None] * len(templates)
    else:
        _masks = masks

    for index, (template, mask) in enumerate(zip(templates, _masks)):
        results = find_all(
            image,
            template,
            mask,
            transparent=transparent,
            threshold=threshold,
            colored=colored,
            remove_duplicate=remove_duplicate,
            debug_output=False,
            preprocessors=preprocessors,
        )
        ret.extend([
            MultipleTemplateMatchResult.from_template_match_result(r, index)
            for r in results
        ])
    # logger.debug(
    #     f'find_all_multi(): templates: {_imgs2str(templates)} images: {_img2str(image)} masks: {_imgs2str(masks)} '
    #     f'result: {_results2str(ret)}'
    # )
    if debug.enabled:
        # 参数表格
        msg = (
            "<center>Templates</center>"
            "<table class='result-table'>"
            "<tr><th>Template</th><th>Mask</th></tr>"
        )
        for t, m in zip(templates, _masks):
            msg += f"<tr><td>{img(t)}</td><td>{img(m)}</td></tr>"
        msg += "</table>"
        msg += "<br>"
        # 结果表格
        msg += (
            "<center>Results</center>"
            "<table class='result-table'>"
            "<tr><th>Template</th><th>Mask</th><th>Result</th></tr>"
        )
        for result in ret:
            template = templates[result.index]
            mask = _masks[result.index]
            msg += f"<tr><td>{img(template)}</td><td>{img(mask)}</td><td>{result.position}</td></tr>"
        msg += "</table>"
        debug_result(
            'image.find_all_multi',
            [_draw_result(image, ret), image], 
            msg
        )
    return ret

def count(
    image: MatLike,
    template: MatLike | str | Image,
    mask: MatLike | str | Image | None = None,
    *,
    transparent: bool = False,
    threshold: float = 0.8,
    remove_duplicate: bool = True,
    colored: bool = False,
    preprocessors: list[PreprocessorProtocol] | None = None,
) -> int:
    """
    指定一个模板，统计其出现的次数。

    :param image: 图像，可以是图像路径或 cv2.Mat。
    :param template: 模板图像，可以是图像路径或 cv2.Mat。
    :param mask: 掩码图像，可以是图像路径或 cv2.Mat。
    :param transparent: 若为 True，则认为输入模板是透明的，并自动将透明模板转换为 Mask 图像。
    :param threshold: 阈值，默认为 0.8。
    :param remove_duplicate: 是否移除重复结果，默认为 True。
    :param colored: 是否匹配颜色，默认为 False。
    :param preprocessors: 预处理列表，默认为 None。
    """
    results = template_match(
        template,
        image,
        mask,
        transparent=transparent,
        threshold=threshold,
        max_results=-1,
        remove_duplicate=remove_duplicate,
        colored=colored,
        preprocessors=preprocessors,
    )
    # logger.debug(
    #     f'count(): template: {_img2str(template)} image: {_img2str(image)} mask: {_img2str(mask)} '
    #     f'result: {_results2str(results)}'
    # )
    if debug.enabled:
        result_image = _draw_result(image, results)
        debug_result(
            'image.count',
            [result_image, image],
            (
                f"template: {img(template)} \n"
                f"mask: {img(mask)} \n"
                f"transparent: {transparent} \n"
                f"threshold: {threshold} \n"
                f"count: {len(results)} \n"
            )
        )
    return len(results)

def expect(
    image: MatLike,
    template: MatLike | str | Image,
    mask: MatLike | str | Image | None = None,
    *,
    transparent: bool = False,
    threshold: float = 0.8,
    colored: bool = False,
    remove_duplicate: bool = True,
    preprocessors: list[PreprocessorProtocol] | None = None,
) -> TemplateMatchResult:
    """
    指定一个模板，寻找其出现的第一个位置。若未找到，则抛出异常。

    :param image: 图像，可以是图像路径或 cv2.Mat。
    :param template: 模板图像，可以是图像路径或 cv2.Mat。
    :param mask: 掩码图像，可以是图像路径或 cv2.Mat。
    :param transparent: 若为 True，则认为输入模板是透明的，并自动将透明模板转换为 Mask 图像。
    :param threshold: 阈值，默认为 0.8。
    :param colored: 是否匹配颜色，默认为 False。
    :param remove_duplicate: 是否移除重复结果，默认为 True。
    :param preprocessors: 预处理列表，默认为 None。
    """
    ret = find(
        image,
        template,
        mask,
        transparent=transparent,
        threshold=threshold,
        colored=colored,
        remove_duplicate=remove_duplicate,
        debug_output=False,
        preprocessors=preprocessors,
    )
    # logger.debug(
    #     f'expect(): template: {_img2str(template)} image: {_img2str(image)} mask: {_img2str(mask)} '
    #     f'result: {_result2str(ret)}'
    # )
    if debug.enabled:
        debug_result(
            'image.expect',
            [_draw_result(image, ret), image],
            (
                f"template: {img(template)} \n"
                f"mask: {img(mask)} \n"
                f"args: transparent={transparent} threshold={threshold} \n"
                f"result: {ret}  "
                '<span class="text-success">SUCCESS</span>' if ret is not None 
                    else '<span class="text-danger">FAILED</span>'
            )
        )
    if ret is None:
        raise TemplateNoMatchError(image, template)
    else:
        return ret

def similar(
    image1: MatLike,
    image2: MatLike,
    threshold: float = 0.9
) -> bool:
    """
    判断两张图像是否相似（灰度）。输入的两张图片必须为相同尺寸。
    """
    if image1.shape != image2.shape:
        raise ValueError('Expected two images with the same size.')
    image1 = cv2.cvtColor(image1, cv2.COLOR_BGR2GRAY)
    image2 = cv2.cvtColor(image2, cv2.COLOR_BGR2GRAY)
    result = structural_similarity(image1, image2, multichannel=True)
    # logger.debug(
    #     f'similar(): image1: {_img2str(image1)} image2: {_img2str(image2)} '
    #     f'result: {result}'
    # )
    # 调试输出
    if debug.enabled:
        result_image = np.hstack([image1, image2])
        debug_result(
            'image.similar',
            [result_image, image1, image2],
            f"result: {result} >= {threshold} == {result >= threshold} \n"
        )
    return result >= threshold





from enum import Enum
from typing import Literal

from .implements.adb import AdbImpl
from .implements.adb_raw import AdbRawImpl
from .implements.windows import WindowsImpl
from .implements.uiautomator2 import UiAutomator2Impl
from .device import Device, AndroidDevice, WindowsDevice

from adbutils import adb

DeviceImpl = Literal['adb', 'adb_raw', 'uiautomator2', 'windows']

def create_device(
    addr: str,
    impl: DeviceImpl,
) -> Device:
    if impl in ['adb', 'adb_raw', 'uiautomator2']:
        result = adb.connect(addr)
        if 'cannot connect to' in result:
            raise ValueError(result)
        d = [d for d in adb.device_list() if d.serial == addr]
        if len(d) == 0:
            raise ValueError(f"Device {addr} not found")
        d = d[0]
        device = AndroidDevice(d)
        if impl == 'adb':
            device._command = AdbImpl(device)
            device._touch = AdbImpl(device)
            device._screenshot = AdbImpl(device)
        elif impl == 'adb_raw':
            device._command = AdbRawImpl(device)
            device._touch = AdbRawImpl(device)
            device._screenshot = AdbRawImpl(device)
        elif impl == 'uiautomator2':
            device._command = UiAutomator2Impl(device)
            device._touch = UiAutomator2Impl(device)
            device._screenshot = UiAutomator2Impl(device)
    elif impl == 'windows':
        device = WindowsDevice()
        device._touch = WindowsImpl(device)
        device._screenshot = WindowsImpl(device)
    return device



import os
import json
import shutil
from importlib import resources
from typing import Literal, Dict, NamedTuple, Tuple, TypeVar, Generic, Any
from typing_extensions import assert_never
from enum import IntEnum, Enum

from pydantic import BaseModel, ConfigDict

# TODO: from kotonebot import config (context) 会和 kotonebot.config 冲突
from kotonebot import logging
from kotonebot.backend.context import config

logger = logging.getLogger(__name__)

T = TypeVar('T')
class ConfigEnum(Enum):
    def display(self) -> str:
        return self.value[1]

class Priority(IntEnum):
    START_GAME = 1
    DEFAULT = 0
    CLAIM_MISSION_REWARD = -1


class APShopItems(IntEnum):
    PRODUCE_PT_UP = 0
    """获取支援强化 Pt 提升"""
    PRODUCE_NOTE_UP = 1
    """获取笔记数提升"""
    RECHALLENGE = 2
    """再挑战券"""
    REGENERATE_MEMORY = 3
    """回忆再生成券"""


倉本千奈_BASE = 0
十王星南_BASE = 100
姫崎莉波_BASE = 200
月村手毬_BASE = 300
有村麻央_BASE = 400
篠泽广_BASE = 500
紫云清夏_BASE = 600
花海佑芽_BASE = 700
花海咲季_BASE = 800
葛城リーリヤ_BASE = 900
藤田ことね_BASE = 1000

class PIdol(IntEnum):
    """P偶像"""
    倉本千奈_Campusmode = 倉本千奈_BASE + 0
    倉本千奈_WonderScale = 倉本千奈_BASE + 1
    倉本千奈_ようこそ初星温泉 = 倉本千奈_BASE + 2
    倉本千奈_仮装狂騒曲 = 倉本千奈_BASE + 3
    倉本千奈_初心 = 倉本千奈_BASE + 4
    倉本千奈_学園生活 = 倉本千奈_BASE + 5
    倉本千奈_日々_発見的ステップ = 倉本千奈_BASE + 6
    倉本千奈_胸を張って一歩ずつ = 倉本千奈_BASE + 7

    十王星南_Campusmode = 十王星南_BASE + 0
    十王星南_一番星 = 十王星南_BASE + 1
    十王星南_学園生活 = 十王星南_BASE + 2
    十王星南_小さな野望 = 十王星南_BASE + 3

    姫崎莉波_clumsytrick = 姫崎莉波_BASE + 0
    姫崎莉波_私らしさのはじまり = 姫崎莉波_BASE + 1
    姫崎莉波_キミとセミブルー = 姫崎莉波_BASE + 2
    姫崎莉波_Campusmode = 姫崎莉波_BASE + 3
    姫崎莉波_LUV = 姫崎莉波_BASE + 4
    姫崎莉波_ようこそ初星温泉 = 姫崎莉波_BASE + 5
    姫崎莉波_ハッピーミルフィーユ = 姫崎莉波_BASE + 6
    姫崎莉波_初心 = 姫崎莉波_BASE + 7
    姫崎莉波_学園生活 = 姫崎莉波_BASE + 8

    月村手毬_Lunasaymaybe = 月村手毬_BASE + 0
    月村手毬_一匹狼 = 月村手毬_BASE + 1
    月村手毬_Campusmode = 月村手毬_BASE + 2
    月村手毬_アイヴイ = 月村手毬_BASE + 3
    月村手毬_初声 = 月村手毬_BASE + 4
    月村手毬_学園生活 = 月村手毬_BASE + 5
    月村手毬_仮装狂騒曲 = 月村手毬_BASE + 6

    有村麻央_Fluorite = 有村麻央_BASE + 0
    有村麻央_はじまりはカッコよく = 有村麻央_BASE + 1
    有村麻央_Campusmode = 有村麻央_BASE + 2
    有村麻央_FeelJewelDream = 有村麻央_BASE + 3
    有村麻央_キミとセミブルー = 有村麻央_BASE + 4
    有村麻央_初恋 = 有村麻央_BASE + 5
    有村麻央_学園生活 = 有村麻央_BASE + 6

    篠泽广_コントラスト = 篠泽广_BASE + 0
    篠泽广_一番向いていないこと = 篠泽广_BASE + 1
    篠泽广_光景 = 篠泽广_BASE + 2
    篠泽广_Campusmode = 篠泽广_BASE + 3
    篠泽广_仮装狂騒曲 = 篠泽广_BASE + 4
    篠泽广_ハッピーミルフィーユ = 篠泽广_BASE + 5
    篠泽广_初恋 = 篠泽广_BASE + 6
    篠泽广_学園生活 = 篠泽广_BASE + 7

    紫云清夏_TameLieOneStep = 紫云清夏_BASE + 0
    紫云清夏_カクシタワタシ = 紫云清夏_BASE + 1
    紫云清夏_夢へのリスタート = 紫云清夏_BASE + 2
    紫云清夏_Campusmode = 紫云清夏_BASE + 3
    紫云清夏_キミとセミブルー = 紫云清夏_BASE + 4
    紫云清夏_初恋 = 紫云清夏_BASE + 5
    紫云清夏_学園生活 = 紫云清夏_BASE + 6
    
    花海佑芽_WhiteNightWhiteWish = 花海佑芽_BASE + 0
    花海佑芽_学園生活 = 花海佑芽_BASE + 1
    花海佑芽_Campusmode = 花海佑芽_BASE + 2
    花海佑芽_TheRollingRiceball = 花海佑芽_BASE + 3
    花海佑芽_アイドル_はじめっ = 花海佑芽_BASE + 4

    花海咲季_BoomBoomPow = 花海咲季_BASE + 0
    花海咲季_Campusmode = 花海咲季_BASE + 1
    花海咲季_FightingMyWay = 花海咲季_BASE + 2
    花海咲季_わたしが一番 = 花海咲季_BASE + 3
    花海咲季_冠菊 = 花海咲季_BASE + 4
    花海咲季_初声 = 花海咲季_BASE + 5
    花海咲季_古今東西ちょちょいのちょい = 花海咲季_BASE + 6
    花海咲季_学園生活 = 花海咲季_BASE + 7

    葛城リーリヤ_一つ踏み出した先に = 葛城リーリヤ_BASE + 0
    葛城リーリヤ_白線 = 葛城リーリヤ_BASE + 1
    葛城リーリヤ_Campusmode = 葛城リーリヤ_BASE + 2
    葛城リーリヤ_WhiteNightWhiteWish = 葛城リーリヤ_BASE + 3
    葛城リーリヤ_冠菊 = 葛城リーリヤ_BASE + 4
    葛城リーリヤ_初心 = 葛城リーリヤ_BASE + 5
    葛城リーリヤ_学園生活 = 葛城リーリヤ_BASE + 6

    藤田ことね_カワイイ_はじめました = 藤田ことね_BASE + 0
    藤田ことね_世界一可愛い私 = 藤田ことね_BASE + 1
    藤田ことね_Campusmode = 藤田ことね_BASE + 2
    藤田ことね_YellowBigBang = 藤田ことね_BASE + 3
    藤田ことね_WhiteNightWhiteWish = 藤田ことね_BASE + 4
    藤田ことね_冠菊 = 藤田ことね_BASE + 5
    藤田ことね_初声 = 藤田ことね_BASE + 6
    藤田ことね_学園生活 = 藤田ことね_BASE + 7

    def to_title(self) -> list[str]:
        match self:
            case PIdol.倉本千奈_Campusmode:
                return ["倉本", "千奈", "Campus", "mode"]
            case PIdol.倉本千奈_WonderScale:
                return ["倉本", "千奈", "Wonder", "Scale"]
            case PIdol.倉本千奈_ようこそ初星温泉:
                return ["倉本", "千奈", "ようこそ初星温泉"]
            case PIdol.倉本千奈_仮装狂騒曲:
                return ["倉本", "千奈", "仮装狂騒曲"]
            case PIdol.倉本千奈_初心:
                return ["倉本", "千奈", "初心"]
            case PIdol.倉本千奈_学園生活:
                return ["倉本", "千奈", "学園生活"]
            case PIdol.倉本千奈_日々_発見的ステップ:
                return ["倉本", "千奈", "日々、発見的ステップ"]
            case PIdol.倉本千奈_胸を張って一歩ずつ:
                return ["倉本", "千奈", "胸を張って一歩ずつ"]
            case PIdol.十王星南_Campusmode:
                return ["十王", "星南", "Campus", "mode"]
            case PIdol.十王星南_一番星:
                return ["十王", "星南", "一番星"]
            case PIdol.十王星南_学園生活:
                return ["十王", "星南", "学園生活"]
            case PIdol.十王星南_小さな野望:
                return ["十王", "星南", "小さな野望"]
            case PIdol.姫崎莉波_clumsytrick:
                return ["姫崎", "莉波", "clumsy", "trick"]
            case PIdol.姫崎莉波_私らしさのはじまり:
                return ["姫崎", "莉波", "『私らしさ』のはじまり"]
            case PIdol.姫崎莉波_キミとセミブルー:
                return ["姫崎", "莉波", "キミとセミブルー"]
            case PIdol.姫崎莉波_Campusmode:
                return ["姫崎", "莉波", "Campus", "mode"]
            case PIdol.姫崎莉波_LUV:
                return ["姫崎", "莉波", "L", "U", "V"]
            case PIdol.姫崎莉波_ようこそ初星温泉:
                return ["姫崎", "莉波", "ようこそ初星温泉"]
            case PIdol.姫崎莉波_ハッピーミルフィーユ:
                return ["姫崎", "莉波", "ハッピーミルフィーユ"]
            case PIdol.姫崎莉波_初心:
                return ["姫崎", "莉波", "初心"]
            case PIdol.姫崎莉波_学園生活:
                return ["姫崎", "莉波", "学園生活"]
            case PIdol.月村手毬_Lunasaymaybe:
                return ["月村", "手毬", "Luna", "say", "maybe"]
            case PIdol.月村手毬_一匹狼:
                return ["月村", "手毬", "一匹狼"]
            case PIdol.月村手毬_Campusmode:
                return ["月村", "手毬", "Campus", "mode"]
            case PIdol.月村手毬_アイヴイ:
                return ["月村", "手毬", "アイヴイ"]
            case PIdol.月村手毬_初声:
                return ["月村", "手毬", "初声"]
            case PIdol.月村手毬_学園生活:
                return ["月村", "手毬", "学園生活"]
            case PIdol.月村手毬_仮装狂騒曲:
                return ["月村", "手毬", "仮装狂騒曲"]
            case PIdol.有村麻央_Fluorite:
                return ["有村", "麻央", "Fluorite"]
            case PIdol.有村麻央_はじまりはカッコよく:
                return ["有村", "麻央", "はじまりはカッコよく"]
            case PIdol.有村麻央_Campusmode:
                return ["有村", "麻央", "Campus", "mode"]
            case PIdol.有村麻央_FeelJewelDream:
                return ["有村", "麻央", "Feel", "Jewel", "Dream"]
            case PIdol.有村麻央_キミとセミブルー:
                return ["有村", "麻央", "キミとセミブルー"]
            case PIdol.有村麻央_初恋:
                return ["有村", "麻央", "初恋"]
            case PIdol.有村麻央_学園生活:
                return ["有村", "麻央", "学園生活"]
            case PIdol.篠泽广_コントラスト:
                return ["篠泽", "広", "コントラスト"]
            case PIdol.篠泽广_一番向いていないこと:
                return ["篠泽", "広", "一番向いていないこと"]
            case PIdol.篠泽广_光景:
                return ["篠泽", "広", "光景"]
            case PIdol.篠泽广_Campusmode:
                return ["篠泽", "広", "Campus", "mode"]
            case PIdol.篠泽广_仮装狂騒曲:
                return ["篠泽", "広", "仮装狂騒曲"]
            case PIdol.篠泽广_ハッピーミルフィーユ:
                return ["篠泽", "広", "ハッピーミルフィーユ"]
            case PIdol.篠泽广_初恋:
                return ["篠泽", "広", "初恋"]
            case PIdol.篠泽广_学園生活:
                return ["篠泽", "広", "学園生活"]
            case PIdol.紫云清夏_TameLieOneStep:
                return ["紫云", "清夏", "Tame", "Lie", "One", "Step"]
            case PIdol.紫云清夏_カクシタワタシ:
                return ["紫云", "清夏", "カクシタワタシ"]
            case PIdol.紫云清夏_夢へのリスタート:
                return ["紫云", "清夏", "夢へのリスタート"]
            case PIdol.紫云清夏_Campusmode:
                return ["紫云", "清夏", "Campus", "mode"]
            case PIdol.紫云清夏_キミとセミブルー:
                return ["紫云", "清夏", "キミとセミブルー"]
            case PIdol.紫云清夏_初恋:
                return ["紫云", "清夏", "初恋"]
            case PIdol.紫云清夏_学園生活:
                return ["紫云", "清夏", "学園生活"]
            case PIdol.花海佑芽_WhiteNightWhiteWish:
                return ["花海", "佑芽", "White", "Night", "Wish"]
            case PIdol.花海佑芽_学園生活:
                return ["花海", "佑芽", "学園生活"]
            case PIdol.花海佑芽_Campusmode:
                return ["花海", "佑芽", "Campus", "mode"]
            case PIdol.花海佑芽_TheRollingRiceball:
                return ["花海", "佑芽", "The", "Rolling", "Riceball"]
            case PIdol.花海佑芽_アイドル_はじめっ:
                return ["花海", "佑芽", "アイドル、はじめっ"]
            case PIdol.花海咲季_BoomBoomPow:
                return ["花海", "咲季", "Boom", "Boom", "Pow"]
            case PIdol.花海咲季_Campusmode:
                return ["花海", "咲季", "Campus", "mode"]
            case PIdol.花海咲季_FightingMyWay:
                return ["花海", "咲季", "Fighting", "My", "Way"]
            case PIdol.花海咲季_わたしが一番:
                return ["花海", "咲季", "わたしが一番"]
            case PIdol.花海咲季_冠菊:
                return ["花海", "咲季", "冠菊"]
            case PIdol.花海咲季_初声:
                return ["花海", "咲季", "初声"]
            case PIdol.花海咲季_古今東西ちょちょいのちょい:
                return ["花海", "咲季", "古今東西ちょちょいのちょい"]
            case PIdol.花海咲季_学園生活:
                return ["花海", "咲季", "学園生活"]
            case PIdol.葛城リーリヤ_一つ踏み出した先に:
                return ["葛城", "リーリヤ", "一つ踏み出した先に"]
            case PIdol.葛城リーリヤ_白線:
                return ["葛城", "リーリヤ", "白線"]
            case PIdol.葛城リーリヤ_Campusmode:
                return ["葛城", "リーリヤ", "Campus", "mode"]
            case PIdol.葛城リーリヤ_WhiteNightWhiteWish:
                return ["葛城", "リーリヤ", "White", "Night", "Wish"]
            case PIdol.葛城リーリヤ_冠菊:
                return ["葛城", "リーリヤ", "冠菊"]
            case PIdol.葛城リーリヤ_初心:
                return ["葛城", "リーリヤ", "初心"]
            case PIdol.葛城リーリヤ_学園生活:
                return ["葛城", "リーリヤ", "学園生活"]
            case PIdol.藤田ことね_カワイイ_はじめました:
                return ["藤田", "ことね", "カワイイ", "はじめました"]
            case PIdol.藤田ことね_世界一可愛い私:
                return ["藤田", "ことね", "世界一可愛い私"]
            case PIdol.藤田ことね_Campusmode:
                return ["藤田", "ことね", "Campus", "mode"]
            case PIdol.藤田ことね_YellowBigBang:
                return ["藤田", "ことね", "Yellow", "Big", "Bang"]
            case PIdol.藤田ことね_WhiteNightWhiteWish:
                return ["藤田", "ことね", "White", "Night", "Wish"]
            case PIdol.藤田ことね_冠菊:
                return ["藤田", "ことね", "冠菊"]
            case PIdol.藤田ことね_初声:
                return ["藤田", "ことね", "初声"]
            case PIdol.藤田ことね_学園生活:
                return ["藤田", "ことね", "学園生活"]
            case _:
                assert_never(self)

class DailyMoneyShopItems(IntEnum):
    """日常商店物品"""
    Recommendations = -1
    """所有推荐商品"""
    LessonNote = 0
    """レッスンノート"""
    VeteranNote = 1
    """ベテランノート"""
    SupportEnhancementPt = 2
    """サポート強化Pt 支援强化Pt"""
    SenseNoteVocal = 3
    """センスノート（ボーカル）感性笔记（声乐）"""
    SenseNoteDance = 4
    """センスノート（ダンス）感性笔记（舞蹈）"""
    SenseNoteVisual = 5
    """センスノート（ビジュアル）感性笔记（形象）"""
    LogicNoteVocal = 6
    """ロジックノート（ボーカル）理性笔记（声乐）"""
    LogicNoteDance = 7
    """ロジックノート（ダンス）理性笔记（舞蹈）"""
    LogicNoteVisual = 8
    """ロジックノート（ビジュアル）理性笔记（形象）"""
    AnomalyNoteVocal = 9
    """アノマリーノート（ボーカル）非凡笔记（声乐）"""
    AnomalyNoteDance = 10
    """アノマリーノート（ダンス）非凡笔记（舞蹈）"""
    AnomalyNoteVisual = 11
    """アノマリーノート（ビジュアル）非凡笔记（形象）"""
    RechallengeTicket = 12
    """再挑戦チケット 重新挑战券"""
    RecordKey = 13
    """記録の鍵 解锁交流的物品"""

    # 碎片
    IdolPiece_倉本千奈_WonderScale = 14
    """倉本千奈 WonderScale 碎片"""
    IdolPiece_篠泽广_光景 = 15
    """篠泽广 光景 碎片"""
    IdolPiece_紫云清夏_TameLieOneStep = 16
    """紫云清夏 Tame-Lie-One-Step 碎片"""
    IdolPiece_葛城リーリヤ_白線 = 17
    """葛城リーリヤ 白線 碎片"""
    IdolPiece_姫崎薪波_cIclumsy_trick = 18
    """姫崎薪波 cIclumsy trick 碎片"""
    IdolPiece_花海咲季_FightingMyWay = 19
    """花海咲季 FightingMyWay 碎片"""
    IdolPiece_藤田ことね_世界一可愛い私 = 20
    """藤田ことね 世界一可愛い私 碎片"""
    IdolPiece_花海佑芽_TheRollingRiceball = 21
    """花海佑芽 The Rolling Riceball 碎片"""
    IdolPiece_月村手毬_LunaSayMaybe = 22
    """月村手毬 Luna say maybe 碎片"""

    @classmethod
    def to_ui_text(cls, item: "DailyMoneyShopItems") -> str:
        """获取枚举值对应的UI显示文本"""
        match item:
            case cls.Recommendations:
                return "所有推荐商品"
            case cls.LessonNote:
                return "课程笔记"
            case cls.VeteranNote:
                return "老手笔记"
            case cls.SupportEnhancementPt:
                return "支援强化点数"
            case cls.SenseNoteVocal:
                return "感性笔记（声乐）"
            case cls.SenseNoteDance:
                return "感性笔记（舞蹈）"
            case cls.SenseNoteVisual:
                return "感性笔记（形象）"
            case cls.LogicNoteVocal:
                return "理性笔记（声乐）"
            case cls.LogicNoteDance:
                return "理性笔记（舞蹈）"
            case cls.LogicNoteVisual:
                return "理性笔记（形象）"
            case cls.AnomalyNoteVocal:
                return "非凡笔记（声乐）"
            case cls.AnomalyNoteDance:
                return "非凡笔记（舞蹈）"
            case cls.AnomalyNoteVisual:
                return "非凡笔记（形象）"
            case cls.RechallengeTicket:
                return "重新挑战券"
            case cls.RecordKey:
                return "记录钥匙"
            case cls.IdolPiece_倉本千奈_WonderScale:
                return "倉本千奈 WonderScale 碎片"
            case cls.IdolPiece_篠泽广_光景:
                return "篠泽广 光景 碎片"
            case cls.IdolPiece_紫云清夏_TameLieOneStep:
                return "紫云清夏 Tame-Lie-One-Step 碎片"
            case cls.IdolPiece_葛城リーリヤ_白線:
                return "葛城リーリヤ 白線 碎片"
            case cls.IdolPiece_姫崎薪波_cIclumsy_trick:
                return "姫崎薪波 cIclumsy trick 碎片"
            case cls.IdolPiece_花海咲季_FightingMyWay:
                return "花海咲季 FightingMyWay 碎片"
            case cls.IdolPiece_藤田ことね_世界一可愛い私:
                return "藤田ことね 世界一可愛い私 碎片"
            case cls.IdolPiece_花海佑芽_TheRollingRiceball:
                return "花海佑芽 The Rolling Riceball 碎片"
            case cls.IdolPiece_月村手毬_LunaSayMaybe:
                return "月村手毬 Luna say maybe 碎片"
            case _:
                assert_never(item)
    
    @classmethod
    def all(cls) -> list[tuple[str, 'DailyMoneyShopItems']]:
        """获取所有枚举值及其对应的UI显示文本"""
        return [(cls.to_ui_text(item), item) for item in cls]
    
    @classmethod
    def _is_note(cls, item: 'DailyMoneyShopItems') -> bool:
        """判断是否为笔记"""
        return 'Note' in item.name and not item.name.startswith('Note') and not item.name.endswith('Note')
    
    @classmethod
    def note_items(cls) -> list[tuple[str, 'DailyMoneyShopItems']]:
        """获取所有枚举值及其对应的UI显示文本"""
        return [(cls.to_ui_text(item), item) for item in cls if cls._is_note(item)]

    def to_resource(self):
        from . import R
        match self:
            case DailyMoneyShopItems.Recommendations:
                return R.Daily.TextShopRecommended
            case DailyMoneyShopItems.LessonNote:
                return R.Shop.ItemLessonNote
            case DailyMoneyShopItems.VeteranNote:
                return R.Shop.ItemVeteranNote
            case DailyMoneyShopItems.SupportEnhancementPt:
                return R.Shop.ItemSupportEnhancementPt
            case DailyMoneyShopItems.SenseNoteVocal:
                return R.Shop.ItemSenseNoteVocal
            case DailyMoneyShopItems.SenseNoteDance:
                return R.Shop.ItemSenseNoteDance
            case DailyMoneyShopItems.SenseNoteVisual:
                return R.Shop.ItemSenseNoteVisual
            case DailyMoneyShopItems.LogicNoteVocal:
                return R.Shop.ItemLogicNoteVocal
            case DailyMoneyShopItems.LogicNoteDance:
                return R.Shop.ItemLogicNoteDance
            case DailyMoneyShopItems.LogicNoteVisual:
                return R.Shop.ItemLogicNoteVisual
            case DailyMoneyShopItems.AnomalyNoteVocal:
                return R.Shop.ItemAnomalyNoteVocal
            case DailyMoneyShopItems.AnomalyNoteDance:
                return R.Shop.ItemAnomalyNoteDance
            case DailyMoneyShopItems.AnomalyNoteVisual:
                return R.Shop.ItemAnomalyNoteVisual
            case DailyMoneyShopItems.RechallengeTicket:
                return R.Shop.ItemRechallengeTicket
            case DailyMoneyShopItems.RecordKey:
                return R.Shop.ItemRecordKey
            case DailyMoneyShopItems.IdolPiece_倉本千奈_WonderScale:
                return R.Shop.IdolPiece.倉本千奈_WonderScale
            case DailyMoneyShopItems.IdolPiece_篠泽广_光景:
                return R.Shop.IdolPiece.篠泽广_光景
            case DailyMoneyShopItems.IdolPiece_紫云清夏_TameLieOneStep:
                return R.Shop.IdolPiece.紫云清夏_TameLieOneStep
            case DailyMoneyShopItems.IdolPiece_葛城リーリヤ_白線:
                return R.Shop.IdolPiece.葛城リーリヤ_白線
            case DailyMoneyShopItems.IdolPiece_姫崎薪波_cIclumsy_trick:
                return R.Shop.IdolPiece.姫崎薪波_cIclumsy_trick
            case DailyMoneyShopItems.IdolPiece_花海咲季_FightingMyWay:
                return R.Shop.IdolPiece.花海咲季_FightingMyWay
            case DailyMoneyShopItems.IdolPiece_藤田ことね_世界一可愛い私:
                return R.Shop.IdolPiece.藤田ことね_世界一可愛い私
            case DailyMoneyShopItems.IdolPiece_花海佑芽_TheRollingRiceball:
                return R.Shop.IdolPiece.花海佑芽_TheRollingRiceball
            case DailyMoneyShopItems.IdolPiece_月村手毬_LunaSayMaybe:
                return R.Shop.IdolPiece.月村手毬_LunaSayMaybe
            case _:
                assert_never(self)

class ConfigBaseModel(BaseModel):
    model_config = ConfigDict(use_attribute_docstrings=True)

class PurchaseConfig(ConfigBaseModel):
    enabled: bool = False
    """是否启用商店购买"""
    money_enabled: bool = False
    """是否启用金币购买"""
    money_items: list[DailyMoneyShopItems] = []
    """金币商店要购买的物品"""
    money_refresh_on: Literal['never', 'not_found', 'always'] = 'never'
    """
    金币商店刷新逻辑。

    * never: 从不刷新。
    * not_found: 仅当要购买的物品不存在时刷新。
    * always: 总是刷新。
    """
    ap_enabled: bool = False
    """是否启用AP购买"""
    ap_items: list[Literal[0, 1, 2, 3]] = []
    """AP商店要购买的物品"""


class ActivityFundsConfig(ConfigBaseModel):
    enabled: bool = False
    """是否启用收取活动费"""


class PresentsConfig(ConfigBaseModel):
    enabled: bool = False
    """是否启用收取礼物"""


class AssignmentConfig(ConfigBaseModel):
    enabled: bool = False
    """是否启用工作"""

    mini_live_reassign_enabled: bool = False
    """是否启用重新分配 MiniLive"""
    mini_live_duration: Literal[4, 6, 12] = 12
    """MiniLive 工作时长"""

    online_live_reassign_enabled: bool = False
    """是否启用重新分配 OnlineLive"""
    online_live_duration: Literal[4, 6, 12] = 12
    """OnlineLive 工作时长"""


class ContestConfig(ConfigBaseModel):
    enabled: bool = False
    """是否启用竞赛"""

    select_which_contestant: Literal[1, 2, 3] = 1
    """选择第几个挑战者"""

class ProduceAction(Enum):
    RECOMMENDED = 'recommended'
    VISUAL = 'visual'
    VOCAL = 'vocal'
    DANCE = 'dance'
    # VISUAL_SP = 'visual_sp'
    # VOCAL_SP = 'vocal_sp'
    # DANCE_SP = 'dance_sp'
    OUTING = 'outing'
    STUDY = 'study'
    ALLOWANCE = 'allowance'
    REST = 'rest'

    @property
    def display_name(self):
        MAP = {
            ProduceAction.RECOMMENDED: '推荐行动',
            ProduceAction.VISUAL: '形象课程',
            ProduceAction.VOCAL: '声乐课程',
            ProduceAction.DANCE: '舞蹈课程',
            ProduceAction.OUTING: '外出（おでかけ）',
            ProduceAction.STUDY: '文化课（授業）',
            ProduceAction.ALLOWANCE: '活动支给（活動支給）',
            ProduceAction.REST: '休息',
        }
        return MAP[self]

class RecommendCardDetectionMode(Enum):
    NORMAL = 'normal'
    STRICT = 'strict'

    @property
    def display_name(self):
        MAP = {
            RecommendCardDetectionMode.NORMAL: '正常模式',
            RecommendCardDetectionMode.STRICT: '严格模式',
        }
        return MAP[self]

class ProduceConfig(ConfigBaseModel):
    enabled: bool = False
    """是否启用培育"""
    mode: Literal['regular', 'pro'] = 'regular'
    """
    培育模式。
    进行一次 REGULAR 培育需要 ~30min，进行一次 PRO 培育需要 ~1h。
    """
    produce_count: int = 1
    """培育的次数。"""
    idols: list[PIdol] = []
    """
    要培育的偶像。将会按顺序循环选择培育。
    若未选择任何偶像，则使用游戏默认选择的偶像（为上次培育偶像）。
    """
    memory_sets: list[int] = []
    """要使用的回忆编成编号，从 1 开始。将会按顺序循环选择使用。"""
    support_card_sets: list[int] = []
    """要使用的支援卡编成编号，从 1 开始。将会按顺序循环选择使用。"""
    auto_set_memory: bool = False
    """是否自动编成回忆。此选项优先级高于回忆编成编号。"""
    auto_set_support_card: bool = False
    """是否自动编成支援卡。此选项优先级高于支援卡编成编号。"""
    use_pt_boost: bool = False
    """是否使用支援强化 Pt 提升。"""
    use_note_boost: bool = False
    """是否使用笔记数提升。"""
    follow_producer: bool = False
    """是否关注租借了支援卡的制作人。"""
    self_study_lesson: Literal['dance', 'visual', 'vocal'] = 'dance'
    """自习课类型。"""
    prefer_lesson_ap: bool = False
    """
    优先 SP 课程。
    
    启用后，若出现 SP 课程，则会优先执行 SP 课程，而不是推荐课程。
    若出现多个 SP 课程，随机选择一个。
    """
    actions_order: list[ProduceAction] = [
        ProduceAction.RECOMMENDED,
        ProduceAction.VISUAL,
        ProduceAction.VOCAL,
        ProduceAction.DANCE,
        ProduceAction.ALLOWANCE,
        ProduceAction.OUTING,
        ProduceAction.STUDY,
        ProduceAction.REST,
    ]
    """
    行动优先级
    
    每一周的行动将会按这里设置的优先级执行。
    """
    recommend_card_detection_mode: RecommendCardDetectionMode = RecommendCardDetectionMode.NORMAL
    """
    推荐卡检测模式
    
    严格模式下，识别速度会降低，但识别准确率会提高。
    """

class MissionRewardConfig(ConfigBaseModel):
    enabled: bool = False
    """是否启用领取任务奖励"""

class ClubRewardConfig(ConfigBaseModel):
    enabled: bool = False
    """是否启用领取社团奖励"""

    selected_note: DailyMoneyShopItems = DailyMoneyShopItems.AnomalyNoteVisual
    """想在社团奖励中获取到的笔记"""

class UpgradeSupportCardConfig(ConfigBaseModel):
    enabled: bool = False
    """是否启用支援卡升级"""

class CapsuleToysConfig(ConfigBaseModel):
    enabled: bool = False
    """是否启用扭蛋机"""

    friend_capsule_toys_count: int = 0
    """好友扭蛋机次数"""

    sense_capsule_toys_count: int = 0
    """感性扭蛋机次数"""

    logic_capsule_toys_count: int = 0
    """理性扭蛋机次数"""

    anomaly_capsule_toys_count: int = 0
    """非凡扭蛋机次数"""
    
class TraceConfig(ConfigBaseModel):
    recommend_card_detection: bool = False
    """跟踪推荐卡检测"""

class StartGameConfig(ConfigBaseModel):
    enabled: bool = True
    """是否启用自动启动游戏。默认为True"""

    start_through_kuyo: bool = False
    """是否通过Kuyo来启动游戏"""

    game_package_name: str = 'com.bandinamcoent.idolmaster_gakuen'
    """游戏包名"""

    kuyo_package_name: str = 'org.kuyo.game'
    """Kuyo包名"""

class BaseConfig(ConfigBaseModel):
    purchase: PurchaseConfig = PurchaseConfig()
    """商店购买配置"""

    activity_funds: ActivityFundsConfig = ActivityFundsConfig()
    """活动费配置"""

    presents: PresentsConfig = PresentsConfig()
    """收取礼物配置"""

    assignment: AssignmentConfig = AssignmentConfig()
    """工作配置"""

    contest: ContestConfig = ContestConfig()
    """竞赛配置"""

    produce: ProduceConfig = ProduceConfig()
    """培育配置"""

    mission_reward: MissionRewardConfig = MissionRewardConfig()
    """领取任务奖励配置"""

    club_reward: ClubRewardConfig = ClubRewardConfig()
    """领取社团奖励配置"""

    upgrade_support_card: UpgradeSupportCardConfig = UpgradeSupportCardConfig()
    """支援卡升级配置"""

    capsule_toys: CapsuleToysConfig = CapsuleToysConfig()
    """扭蛋机配置"""

    trace: TraceConfig = TraceConfig()
    """跟踪配置"""

    start_game: StartGameConfig = StartGameConfig()
    """启动游戏配置"""


def conf() -> BaseConfig:
    """获取当前配置数据"""
    c = config.to(BaseConfig).current
    return c.options

def sprite_path(path: str) -> str:
    standalone = os.path.join('kotonebot/tasks/sprites', path)
    if os.path.exists(standalone):
        return standalone
    return str(resources.files('kotonebot.tasks.sprites') / path)

def upgrade_config() -> str | None:
    """
    升级配置文件
    """
    if not os.path.exists('config.json'):
        return None
    with open('config.json', 'r', encoding='utf-8') as f:
        root = json.load(f)
    
    user_configs = root['user_configs']
    old_version = root['version']
    messages = []
    def upgrade_user_config(version: int, user_config: dict[str, Any]) -> int:
        nonlocal messages
        while True:
            match version:
                case 1:
                    logger.info('Upgrading config: v1 -> v2')
                    user_config, msg = upgrade_v1_to_v2(user_config['options'])
                    messages.append(msg)
                    version = 2
                case _:
                    logger.info('No config upgrade needed.')
                    return version
    for user_config in user_configs:
        new_version = upgrade_user_config(old_version, user_config)
        root['version'] = new_version

    with open('config.json', 'w', encoding='utf-8') as f:
        json.dump(root, f, ensure_ascii=False, indent=4)
    return '\n'.join(messages)
    
def upgrade_v1_to_v2(options: dict[str, Any]) -> tuple[dict[str, Any], str | None]:
    """
    v1 -> v2 变更：

    1. 将 PIdol 的枚举值改为整数
    """
    msg = ''
    # 转换 PIdol 的枚举值
    def map_idol(idol: list[str]) -> PIdol | None:
        logger.debug("Converting %s", idol)
        match idol:
            case ["倉本千奈", "Campus mode!!"]:
                return PIdol.倉本千奈_Campusmode
            case ["倉本千奈", "Wonder Scale"]:
                return PIdol.倉本千奈_WonderScale
            case ["倉本千奈", "ようこそ初星温泉"]:
                return PIdol.倉本千奈_ようこそ初星温泉
            case ["倉本千奈", "仮装狂騒曲"]:
                return PIdol.倉本千奈_仮装狂騒曲
            case ["倉本千奈", "初心"]:
                return PIdol.倉本千奈_初心
            case ["倉本千奈", "学園生活"]:
                return PIdol.倉本千奈_学園生活
            case ["倉本千奈", "日々、発見的ステップ！"]:
                return PIdol.倉本千奈_日々_発見的ステップ
            case ["倉本千奈", "胸を張って一歩ずつ"]:
                return PIdol.倉本千奈_胸を張って一歩ずつ
            case ["十王星南", "Campus mode!!"]:
                return PIdol.十王星南_Campusmode
            case ["十王星南", "一番星"]:
                return PIdol.十王星南_一番星
            case ["十王星南", "学園生活"]:
                return PIdol.十王星南_学園生活
            case ["十王星南", "小さな野望"]:
                return PIdol.十王星南_小さな野望
            case ["姫崎莉波", "clumsy trick"]:
                return PIdol.姫崎莉波_clumsytrick
            case ["姫崎莉波", "『私らしさ』のはじまり"]:
                return PIdol.姫崎莉波_私らしさのはじまり
            case ["姫崎莉波", "キミとセミブルー"]:
                return PIdol.姫崎莉波_キミとセミブルー
            case ["姫崎莉波", "Campus mode!!"]:
                return PIdol.姫崎莉波_Campusmode
            case ["姫崎莉波", "L.U.V"]:
                return PIdol.姫崎莉波_LUV
            case ["姫崎莉波", "ようこそ初星温泉"]:
                return PIdol.姫崎莉波_ようこそ初星温泉
            case ["姫崎莉波", "ハッピーミルフィーユ"]:
                return PIdol.姫崎莉波_ハッピーミルフィーユ
            case ["姫崎莉波", "初心"]:
                return PIdol.姫崎莉波_初心
            case ["姫崎莉波", "学園生活"]:
                return PIdol.姫崎莉波_学園生活
            case ["月村手毬", "Luna say maybe"]:
                return PIdol.月村手毬_Lunasaymaybe
            case ["月村手毬", "一匹狼"]:
                return PIdol.月村手毬_一匹狼
            case ["月村手毬", "Campus mode!!"]:
                return PIdol.月村手毬_Campusmode
            case ["月村手毬", "アイヴイ"]:
                return PIdol.月村手毬_アイヴイ
            case ["月村手毬", "初声"]:
                return PIdol.月村手毬_初声
            case ["月村手毬", "学園生活"]:
                return PIdol.月村手毬_学園生活
            case ["月村手毬", "仮装狂騒曲"]:
                return PIdol.月村手毬_仮装狂騒曲
            case ["有村麻央", "Fluorite"]:
                return PIdol.有村麻央_Fluorite
            case ["有村麻央", "はじまりはカッコよく"]:
                return PIdol.有村麻央_はじまりはカッコよく
            case ["有村麻央", "Campus mode!!"]:
                return PIdol.有村麻央_Campusmode
            case ["有村麻央", "Feel Jewel Dream"]:
                return PIdol.有村麻央_FeelJewelDream
            case ["有村麻央", "キミとセミブルー"]:
                return PIdol.有村麻央_キミとセミブルー
            case ["有村麻央", "初恋"]:
                return PIdol.有村麻央_初恋
            case ["有村麻央", "学園生活"]:
                return PIdol.有村麻央_学園生活
            case ["篠泽广", "コントラスト"]:
                return PIdol.篠泽广_コントラスト
            case ["篠泽广", "一番向いていないこと"]:
                return PIdol.篠泽广_一番向いていないこと
            case ["篠泽广", "光景"]:
                return PIdol.篠泽广_光景
            case ["篠泽广", "Campus mode!!"]:
                return PIdol.篠泽广_Campusmode
            case ["篠泽广", "仮装狂騒曲"]:
                return PIdol.篠泽广_仮装狂騒曲
            case ["篠泽广", "ハッピーミルフィーユ"]:
                return PIdol.篠泽广_ハッピーミルフィーユ
            case ["篠泽广", "初恋"]:
                return PIdol.篠泽广_初恋
            case ["篠泽广", "学園生活"]:
                return PIdol.篠泽广_学園生活
            case ["紫云清夏", "Tame-Lie-One-Step"]:
                return PIdol.紫云清夏_TameLieOneStep
            case ["紫云清夏", "カクシタワタシ"]:
                return PIdol.紫云清夏_カクシタワタシ
            case ["紫云清夏", "夢へのリスタート"]:
                return PIdol.紫云清夏_夢へのリスタート
            case ["紫云清夏", "Campus mode!!"]:
                return PIdol.紫云清夏_Campusmode
            case ["紫云清夏", "キミとセミブルー"]:
                return PIdol.紫云清夏_キミとセミブルー
            case ["紫云清夏", "初恋"]:
                return PIdol.紫云清夏_初恋
            case ["紫云清夏", "学園生活"]:
                return PIdol.紫云清夏_学園生活
            case ["花海佑芽", "White Night! White Wish!"]:
                return PIdol.花海佑芽_WhiteNightWhiteWish
            case ["花海佑芽", "学園生活"]:
                return PIdol.花海佑芽_学園生活
            case ["花海佑芽", "Campus mode!!"]:
                return PIdol.花海佑芽_Campusmode
            case ["花海佑芽", "The Rolling Riceball"]:
                return PIdol.花海佑芽_TheRollingRiceball
            case ["花海佑芽", "アイドル、はじめっ！"]:
                return PIdol.花海佑芽_アイドル_はじめっ
            case ["花海咲季", "Boom Boom Pow"]:
                return PIdol.花海咲季_BoomBoomPow
            case ["花海咲季", "Campus mode!!"]:
                return PIdol.花海咲季_Campusmode
            case ["花海咲季", "Fighting My Way"]:
                return PIdol.花海咲季_FightingMyWay
            case ["花海咲季", "わたしが一番！"]:
                return PIdol.花海咲季_わたしが一番
            case ["花海咲季", "冠菊"]:
                return PIdol.花海咲季_冠菊
            case ["花海咲季", "初声"]:
                return PIdol.花海咲季_初声
            case ["花海咲季", "古今東西ちょちょいのちょい"]:
                return PIdol.花海咲季_古今東西ちょちょいのちょい
            case ["花海咲季", "学園生活"]:
                return PIdol.花海咲季_学園生活
            case ["葛城リーリヤ", "一つ踏み出した先に"]:
                return PIdol.葛城リーリヤ_一つ踏み出した先に
            case ["葛城リーリヤ", "白線"]:
                return PIdol.葛城リーリヤ_白線
            case ["葛城リーリヤ", "Campus mode!!"]:
                return PIdol.葛城リーリヤ_Campusmode
            case ["葛城リーリヤ", "White Night! White Wish!"]:
                return PIdol.葛城リーリヤ_WhiteNightWhiteWish
            case ["葛城リーリヤ", "冠菊"]:
                return PIdol.葛城リーリヤ_冠菊
            case ["葛城リーリヤ", "初心"]:
                return PIdol.葛城リーリヤ_初心
            case ["葛城リーリヤ", "学園生活"]:
                return PIdol.葛城リーリヤ_学園生活
            case ["藤田ことね", "カワイイ", "はじめました"]:
                return PIdol.藤田ことね_カワイイ_はじめました
            case ["藤田ことね", "世界一可愛い私"]:
                return PIdol.藤田ことね_世界一可愛い私
            case ["藤田ことね", "Campus mode!!"]:
                return PIdol.藤田ことね_Campusmode
            case ["藤田ことね", "Yellow Big Bang！"]:
                return PIdol.藤田ことね_YellowBigBang
            case ["藤田ことね", "White Night! White Wish!"]:
                return PIdol.藤田ことね_WhiteNightWhiteWish
            case ["藤田ことね", "冠菊"]:
                return PIdol.藤田ことね_冠菊
            case ["藤田ことね", "初声"]:
                return PIdol.藤田ことね_初声
            case ["藤田ことね", "学園生活"]:
                return PIdol.藤田ことね_学園生活
            case _:
                nonlocal msg
                if msg == '':
                    msg = '培育设置中的以下偶像升级失败。请尝试手动添加。\n'
                msg += f'{idol} 未找到\n'
                return None
    old_idols = options['produce']['idols']
    new_idols = list(filter(lambda x: x is not None, map(map_idol, old_idols)))
    options['produce']['idols'] = new_idols
    shutil.copy('config.json', 'config.v1.json')
    return options, msg


if __name__ == '__main__':
    print(PurchaseConfig.model_fields['money_refresh_on'].description)


import time
import socket
from typing import Protocol, NamedTuple
from dataclasses import dataclass

from adbutils import adb, AdbTimeout, AdbError
from adbutils._device import AdbDevice

from kotonebot import logging
from kotonebot.util import Countdown, Interval

logger = logging.getLogger(__name__)

def tcp_ping(host: str, port: int, timeout: float = 1.0) -> bool:
    """
    通过 TCP ping 检查主机和端口是否可达。
    
    :param host: 主机名或 IP 地址
    :param port: 端口号
    :param timeout: 超时时间（秒）
    :return: 如果主机和端口可达，则返回 True，否则返回 False
    """
    logger.debug('TCP ping %s:%d...', host, port)
    try:
        with socket.create_connection((host, port), timeout):
            logger.debug('TCP ping %s:%d success.', host, port)
            return True
    except (socket.timeout, ConnectionRefusedError, OSError):
        logger.debug('TCP ping %s:%d failed.', host, port)
        return False


@dataclass
class Instance:
    id: str
    name: str
    adb_port: int
    adb_ip: str = '127.0.0.1'
    adb_emulator_name: str = 'emulator-5554'

    def start(self):
        raise NotImplementedError()
    
    def stop(self):
        raise NotImplementedError()

    def running(self) -> bool:
        raise NotImplementedError()

    def wait_available(self, timeout: float = 180):
        logger.info('Starting to wait for emulator %s(127.0.0.1:%d) to be available...', self.name, self.adb_port)
        state = 0
        port = self.adb_port
        emulator_name = self.adb_emulator_name
        cd = Countdown(timeout)
        it = Interval(1)
        d: AdbDevice | None = None
        while True:
            if cd.expired():
                raise TimeoutError(f"Emulator {self.name} is not available.")
            it.wait()
            try:
                match state:
                    case 0:
                        logger.debug('Ping emulator %s(127.0.0.1:%d)...', self.name, port)
                        if tcp_ping('127.0.0.1', port):
                            logger.debug('Ping emulator %s(127.0.0.1:%d) success.', self.name, port)
                            state = 1
                    case 1:
                        logger.debug('Connecting to emulator %s(127.0.0.1:%d)...', self.name, port)
                        if adb.connect(f'127.0.0.1:{port}', timeout=0.5):
                            logger.debug('Connect to emulator %s(127.0.0.1:%d) success.', self.name, port)
                            state = 2
                    case 2:
                        logger.debug('Getting device list...')
                        if devices := adb.device_list():
                            logger.debug('Get device list success. devices=%s', devices)
                            # emulator_name 用于适配雷电模拟器
                            # 雷电模拟器启动后，在上方的列表中并不会出现 127.0.0.1:5555，而是 emulator-5554
                            d = next(
                                (d for d in devices if d.serial == f'127.0.0.1:{port}' or d.serial == emulator_name),
                                None
                            )
                            if d:
                                logger.debug('Get target device success. d=%s', d)
                                state = 3
                    case 3:
                        if not d:
                            logger.warning('Device is None.')
                            state = 0
                            continue
                        logger.debug('Waiting for device state...')
                        if d.get_state() == 'device':
                            logger.debug('Device state ready. state=%s', d.get_state())
                            state = 4
                    case 4:
                        logger.debug('Waiting for device boot completed...')
                        if not d:
                            logger.warning('Device is None.')
                            state = 0
                            continue
                        ret = d.shell('getprop sys.boot_completed')
                        if isinstance(ret, str) and ret.strip() == '1':
                            logger.debug('Device boot completed. ret=%s', ret)
                            state = 5
                    case 5:
                        if not d:
                            logger.warning('Device is None.')
                            state = 0
                            continue
                        app = d.app_current()
                        logger.debug('Waiting for launcher... (current=%s)', app)
                        if app and 'launcher' in app.package:
                            logger.info('Emulator %s(127.0.0.1:%d) now is available.', self.name, self.adb_port)
                            state = 6
                    case 6:
                        break
            except (AdbError, AdbTimeout):
                state = 1
                continue
        time.sleep(1)
        logger.info('Emulator %s(127.0.0.1:%d) now is available.', self.name, self.adb_port)


class HostProtocol(Protocol):
    @staticmethod
    def installed() -> bool: ...
    @staticmethod
    def list() -> list[Instance]: ...


if __name__ == '__main__':
    from . import bluestack_global
    from pprint import pprint
    logging.basicConfig(level=logging.DEBUG, format='[%(asctime)s][%(levelname)s] %(message)s')
    # bluestack_global.
    ins = Instance(id='1', name='test', adb_port=5555)
    ins.wait_available()

    # 
    # while not tcp_ping('127.0.0.1', 16384):
    #     print('waiting for bluestacks to start...')
    
    # while True:
    #     print('connecting to bluestacks...')
    #     try:
    #         adb.connect('127.0.0.1:16384', timeout=0.1)
    #         print('connected to bluestacks')
    #         if d := adb.device_list()[0]:
    #             if d.get_state() == 'device':
    #                 if d.shell('getprop sys.boot_completed').strip() == '1':
    #                     if 'launcher' in d.app_current().package:
    #                         break
    #     except Exception as e:
    #         print(e)
    #     time.sleep(0.5)
    # time.sleep(1)
    # print('bluestacks is ready')



import os
import json
from typing import Sequence
import requests
from cv2.typing import MatLike
from dotenv import dotenv_values

from .image_host import upload
from .protocol import PushkitProtocol

config = dotenv_values(".env")

class Wxpusher(PushkitProtocol):
    def __init__(self, app_token: str | None = None, uid: str | None = None):
        self.app_token = app_token or config["WXPUSHER_APP_TOKEN"]
        self.uid = uid or config["WXPUSHER_UID"]

    def push(self, title: str, message: str, *, images: Sequence[str | MatLike] | None = None) -> None:
        summary = title
        content = message

        if images:
            image_urls = upload(images)
            img_md = "\n".join([f"![{img_url}]({img_url})" for img_url in image_urls])
            content = content + "\n" + img_md

        data = {
            "appToken": self.app_token,
            "uid": self.uid,
            "summary": summary,
            "content": content,
            "contentType": 3, # 1: 文本 2: HTML 3: Markdown
            "uids": [self.uid],
            "verifyPay": False,
            "verifyPayType": 0
        }

        response = requests.post(
            "http://wxpusher.zjiecode.com/api/send/message",
            json=data
        )
        result = response.json()
        
        if result["code"] != 1000 or not result["success"]:
            raise RuntimeError(f"推送失败: {result['msg']}")

# TODO: 极简推送 https://wxpusher.zjiecode.com/docs/#/?id=spt
        
if __name__ == "__main__":
    import cv2
    wxpusher = Wxpusher()
    wxpusher.push("测试图片", "测试图片", images=[cv2.imread("res/sprites/jp/common/button_close.png")])




"""启动游戏，领取登录奖励，直到首页为止"""
import os
import logging

from . import R
from .common import Priority, conf
from .actions.loading import loading
from kotonebot.util import Countdown, Interval
from .actions.scenes import at_home, goto_home
from .actions.commu import handle_unread_commu
from kotonebot.errors import GameUpdateNeededError
from kotonebot import task, action, device, image, ocr

logger = logging.getLogger(__name__)

@action('启动游戏.进入首页', screenshot_mode='manual-inherit')
def wait_for_home():
    """
    前置条件：游戏已启动\n
    结束状态：游戏首页
    """
    logger.info('Entering home...')
    it = Interval()
    click_cd = Countdown(1).start()
    should_click = False
    while True:
        device.screenshot()
        # 首页
        if image.find(R.Daily.ButtonHomeCurrent):
            break
        # TAP TO START 画面
        # [screenshots/startup/1.png]
        elif image.find(R.Daily.ButonLinkData):
            should_click = True
        elif loading():
            pass
        # 热更新
        # [screenshots/startup/update.png]
        elif image.find(R.Common.TextGameUpdate):
            device.click(image.expect(R.Common.ButtonConfirm))
        # 本体更新
        # [kotonebot-resource/sprites/jp/daily/screenshot_apk_update.png]
        elif ocr.find('アップデート', rect=R.Daily.BoxApkUpdateDialogTitle):
            raise GameUpdateNeededError()
        # 公告
        # [screenshots/startup/announcement1.png]
        elif image.find(R.Common.ButtonIconClose):
            device.click()
        # 生日
        # [screenshots/startup/birthday.png]
        elif handle_unread_commu():
            pass

        if should_click and click_cd.expired():
            device.click_center()
            click_cd.reset()
        it.wait()

@action('启动游戏.Android', screenshot_mode='manual-inherit')
def android_launch():
    """
    前置条件：-
    结束状态：-
    """
    # 如果已经在游戏中，直接返回home
    if device.current_package() == conf().start_game.game_package_name:
        logger.info("Game already started")
        if not at_home():
            logger.info("Not at home, going to home")
            goto_home()
        return
    
    # 如果不在游戏中，启动游戏
    if not conf().start_game.start_through_kuyo:
        # 直接启动
        device.launch_app(conf().start_game.game_package_name)
    else:
        # 通过Kuyo启动
        if device.current_package() == conf().start_game.kuyo_package_name:
            logger.warning("Kuyo already started. Auto start game failed.")
            # TODO: Kuyo支持改进
            return
        # 启动kuyo
        device.launch_app('org.kuyo.game')
        # 点击"加速"
        device.click(image.expect_wait(R.Kuyo.ButtonTab3Speedup, timeout=10))
        # 点击"K空间启动"
        device.click(image.expect_wait(R.Kuyo.ButtonStartGame, timeout=10))

@action('启动游戏.Windows', screenshot_mode='manual-inherit')
def windows_launch():
    """
    前置条件：-
    结束状态：游戏窗口出现
    """
    from ahk import AHK
    from importlib import resources

    ahk_path = str(resources.files('kaa.res.bin') / 'AutoHotkey.exe')
    ahk = AHK(executable_path=ahk_path)
    logger.info('Starting game...')
    os.startfile('dmmgameplayer://play/GCL/gakumas/cl/win')
    # 等待游戏窗口出现
    it = Interval()
    while True:
        if ahk.find_window(title='gakumas'):
            logger.debug('Game window found.')
            break
        logger.debug('Waiting for game window...')
        it.wait()

@task('启动游戏', priority=Priority.START_GAME)
def start_game():
    """
    启动游戏，直到游戏进入首页为止。
    """
    if not conf().start_game.enabled:
        logger.info('"Start game" is disabled.')
        return
    
    if device.platform == 'android':
        android_launch()
    elif device.platform == 'windows':
        windows_launch()
    else:
        raise ValueError(f'Unsupported platform: {device.platform}')

    wait_for_home()

if __name__ == '__main__':
    import logging
    logging.basicConfig(level=logging.INFO, format='[%(asctime)s] [%(levelname)s] [%(name)s] [%(funcName)s] [%(lineno)d] %(message)s')
    logger.setLevel(logging.DEBUG)
    start_game()




from typing import Protocol

from cv2.typing import MatLike

class PushkitProtocol(Protocol):
    def push(
        self,
        title: str,
        message: str,
        *,
        images: list[str | MatLike] | None = None,
    ) -> None:
        ...



from typing import Literal
from logging import getLogger

from kotonebot.backend.preprocessor import HsvColorFilter
from kotonebot.tasks.actions.loading import loading

from .. import R
from kotonebot import (
    ocr,
    device,
    contains,
    image,
    regex,
    action,
    sleep,
    Interval,
)
from ..game_ui import CommuEventButtonUI, WhiteFilter
from .pdorinku import acquire_pdorinku
from kotonebot.backend.dispatch import SimpleDispatcher
from kotonebot.tasks.actions.commu import handle_unread_commu
from kotonebot.util import measure_time

logger = getLogger(__name__)

@action('领取技能卡', screenshot_mode='manual-inherit')
def acquire_skill_card():
    """获取技能卡（スキルカード）"""
    # TODO: 识别卡片内容，而不是固定选卡
    # TODO: 不硬编码坐标
    logger.debug("Locating all skill cards...")
    device.screenshot()
    cards = image.find_all_multi([
        R.InPurodyuusu.A,
        R.InPurodyuusu.M
    ])
    cards = sorted(cards, key=lambda x: (x.position[0], x.position[1]))
    logger.info(f"Found {len(cards)} skill cards")
    logger.debug("Click first skill card")
    device.click(cards[0].rect)
    sleep(0.2)
    logger.debug("Click acquire button")
    device.click(image.expect_wait(R.InPurodyuusu.AcquireBtnDisabled))
    # acquisitions(['PSkillCardSelect']) 优先做这个
    # device.screenshot()
    # (SimpleDispatcher('acquire_skill_card')
    #     .click(contains("受け取る"), finish=True,  log="Skill card #1 acquired")
    #     # .click_any([
    #     #     R.InPurodyuusu.PSkillCardIconBlue,
    #     #     R.InPurodyuusu.PSkillCardIconColorful
    #     # ], finish=True, log="Skill card #1 acquired")
    # ).run()
    # # logger.info("Skill card #1 acquired")

@action('选择P物品', screenshot_mode='auto')
def select_p_item():
    """
    前置条件：P物品选择对话框（受け取るＰアイテムを選んでください;）\n
    结束状态：P物品获取动画
    """
    # 前置条件 [screenshots/produce/in_produce/select_p_item.png]
    # 前置条件 [screenshots/produce/in_produce/claim_p_item.png]

    POSTIONS = [
        (157, 820, 128, 128), # x, y, w, h
        (296, 820, 128, 128),
        (435, 820, 128, 128),
    ] # TODO: HARD CODED
    device.click(POSTIONS[0])
    sleep(0.5)
    device.click(ocr.expect_wait('受け取る'))

@action('技能卡自选强化', screenshot_mode='manual-inherit')
def hanlde_skill_card_enhance():
    """
    前置条件：技能卡强化对话框\n
    结束状态：技能卡强化动画结束后瞬间

    :return: 是否成功处理对话框
    """
    # 前置条件 [kotonebot-resource\sprites\jp\in_purodyuusu\screenshot_skill_card_enhane.png]
    # 结束状态 [screenshots/produce/in_produce/skill_card_enhance.png]
    cards = image.find_multi([
        R.InPurodyuusu.A,
        R.InPurodyuusu.M
    ])
    if cards is None:
        logger.info("No skill cards found")
        return False
    logger.debug("Clicking first skill card.")
    device.click(cards)
    it = Interval()
    while True:
        device.screenshot()
        if image.find(R.InPurodyuusu.ButtonEnhance, colored=True):
            device.click()
            logger.debug("Enhance button found")
        elif image.find(R.InPurodyuusu.IconSkillCardEventBubble):
            device.click_center()
            logger.debug("Skill card event bubble found")
            break
        it.wait()
    logger.debug("Handle skill card enhance finished.")

@action('技能卡自选删除', screenshot_mode='manual-inherit')
def handle_skill_card_removal():
    """
    前置条件：技能卡删除对话框\n
    结束状态：技能卡删除动画结束后瞬间
    """
    # 前置条件 [kotonebot-resource\sprites\jp\in_purodyuusu\screenshot_remove_skill_card.png]
    card = image.find_multi([
        R.InPurodyuusu.A,
        R.InPurodyuusu.M
    ])
    if card is None:
        logger.info("No skill cards found")
        return False
    device.click(card)
    it = Interval()
    while True:
        device.screenshot()
        if image.find(R.InPurodyuusu.ButtonRemove):
            device.click()
            logger.debug("Remove button clicked.")
        elif image.find(R.InPurodyuusu.IconSkillCardEventBubble):
            device.click_center()
            logger.debug("Skill card event bubble found")
            break
        it.wait()
    logger.debug("Handle skill card removal finished.")

AcquisitionType = Literal[
    "PDrinkAcquire", # P饮料被动领取
    "PDrinkSelect", # P饮料主动领取
    "PDrinkMax", # P饮料到达上限
    "PSkillCardAcquire", # 技能卡领取
    "PSkillCardSelect", # 技能卡选择
    "PSkillCardEnhanced", # 技能卡强化
    "PSkillCardEnhanceSelect", # 技能卡自选强化
    "PSkillCardRemoveSelect", # 技能卡自选删除
    "PSkillCardEvent", # 技能卡事件（随机强化、删除、更换）
    "PItemClaim", # P物品领取
    "PItemSelect", # P物品选择
    "Clear", # 目标达成
    "ClearNext", # 目标达成 NEXT
    "NetworkError", # 网络中断弹窗
    "SkipCommu", # 跳过交流
    "Loading", # 加载画面
]

@measure_time()
@action('处理培育事件', screenshot_mode='manual')
def acquisitions() -> AcquisitionType | None:
    """处理行动开始前和结束后可能需要处理的事件，直到到行动页面为止"""
    img = device.screenshot()

    screen_size = device.screen_size
    bottom_pos = (int(screen_size[0] * 0.5), int(screen_size[1] * 0.7)) # 底部中间
    logger.info("Acquisition stuffs...")

    # 加载画面
    if loading():
        logger.info("Loading...")
        return "Loading"

    # P饮料领取
    logger.debug("Check PDrink acquire...")
    if image.find(R.InPurodyuusu.PDrinkIcon):
        logger.info("PDrink acquire found")
        device.click_center()
        sleep(1)
        return "PDrinkAcquire"
    # P饮料到达上限
    logger.debug("Check PDrink max...")
    # TODO: 需要封装一个更好的实现方式。比如 wait_stable？
    if image.find(R.InPurodyuusu.TextPDrinkMax):
        logger.debug("PDrink max found")
        device.screenshot()
        if image.find(R.InPurodyuusu.TextPDrinkMax):
            # 有对话框标题，但是没找到确认按钮
            # 可能是需要勾选一个饮料
            if not image.find(R.InPurodyuusu.ButtonLeave, colored=True):
                logger.info("No leave button found, click checkbox")
                device.click(image.expect(R.Common.CheckboxUnchecked, colored=True))
                sleep(0.2)
                device.screenshot()
            if leave := image.find(R.InPurodyuusu.ButtonLeave, colored=True):
                logger.info("Leave button found")
                device.click(leave)
                return "PDrinkMax"
    if image.find(R.InPurodyuusu.TextPDrinkMaxConfirmTitle):
        logger.debug("PDrink max confirm found")
        device.screenshot()
        if image.find(R.InPurodyuusu.TextPDrinkMaxConfirmTitle):
            if confirm := image.find(R.Common.ButtonConfirm):
                logger.info("Confirm button found")
                device.click(confirm)
                return "PDrinkMax"
    # 技能卡领取
    logger.debug("Check skill card acquisition...")
    if image.find_multi([
        R.InPurodyuusu.PSkillCardIconBlue,
        R.InPurodyuusu.PSkillCardIconColorful
    ]):
        logger.info("Acquire skill card found")
        device.click_center()
        return "PSkillCardAcquire"

    # 技能卡自选强化
    if image.find(R.InPurodyuusu.IconTitleSkillCardEnhance):
        if hanlde_skill_card_enhance():
            return "PSkillCardEnhanceSelect"

    # 技能卡自选删除
    if image.find(R.InPurodyuusu.IconTitleSkillCardRemoval):
        if handle_skill_card_removal():
            return "PSkillCardRemoveSelect"

    # 目标达成
    logger.debug("Check gloal clear (達成)...")
    if image.find(R.InPurodyuusu.IconClearBlue):
        logger.info("Clear found")
        logger.debug("Goal clear (達成): clicked")
        device.click_center()
        sleep(1)
        return "Clear"
    # 目标达成 NEXT
    if image.find(R.InPurodyuusu.TextGoalClearNext, preprocessors=[WhiteFilter()]):
        logger.info("Goal clear (達成) next found")
        device.click_center()
        sleep(1)
        return "ClearNext"
    # P物品领取
    logger.debug("Check PItem claim...")
    if image.find(R.InPurodyuusu.PItemIconColorful):
        logger.info("Click to finish PItem acquisition")
        device.click_center()
        sleep(1)
        return "PItemClaim"

    # 网络中断弹窗
    logger.debug("Check network error popup...")
    if image.find(R.Common.TextNetworkError):
        logger.info("Network error popup found")
        device.click(image.expect(R.Common.ButtonRetry))
        return "NetworkError"
    # 跳过未读交流
    logger.debug("Check skip commu...")
    if handle_unread_commu(img):
        return "SkipCommu"

    # === 需要 OCR 的放在最后执行 ===

    # 物品选择对话框
    logger.debug("Check award select dialog...")
    if image.find(R.InPurodyuusu.TextClaim):
        logger.info("Award select dialog found.")

        # P饮料选择
        logger.debug("Check PDrink select...")
        if image.find(R.InPurodyuusu.TextPDrink):
            logger.info("PDrink select found")
            acquire_pdorinku(index=0)
            return "PDrinkSelect"
        # 技能卡选择
        logger.debug("Check skill card select...")
        if image.find(R.InPurodyuusu.TextSkillCard):
            logger.info("Acquire skill card found")
            acquire_skill_card()
            return "PSkillCardSelect"
        # P物品选择
        logger.debug("Check PItem select...")
        if image.find(R.InPurodyuusu.TextPItem):
            logger.info("Acquire PItem found")
            select_p_item()
            return "PItemSelect"

    # 技能卡变更事件
    # 包括下面这些：
    # 1. 技能卡更换
    # [screenshots/produce/in_produce/support_card_change.png]
    # 2. 技能卡强化
    # [screenshots/produce/in_produce/skill_card_enhance.png]
    # 3. 技能卡移除
    # [screenshots/produce/in_produce/skill_card_removal.png]
    logger.debug("Check skill card events...")
    if image.find(R.InPurodyuusu.IconSkillCardEventBubble):
        device.click() # 不能 click_center，因为中间是技能卡
        return "PSkillCardEvent"
    
    # 技能卡获取
    # [kotonebot-resource\sprites\jp\in_purodyuusu\screenshot_skill_card_acquired.png]
    # 因为这个文本有缩放动画，因此暂时没法用模板匹配代替
    if ocr.find("スキルカード獲得", rect=R.InPurodyuusu.BoxSkillCardAcquired):
        logger.info("Acquire skill card from loot box")
        device.click_center()
        # 下面就是普通的技能卡选择
        sleep(0.2)
        return acquisitions()

    return None

def until_acquisition_clear():
    """
    处理各种奖励、弹窗，直到没有新的奖励、弹窗为止

    前置条件：任意\n
    结束条件：任意
    """
    interval = Interval(0.6)
    while acquisitions():
        interval.wait()

@action('处理交流事件', screenshot_mode='manual-inherit')
def commut_event():
    ui = CommuEventButtonUI()
    buttons = ui.all(description=False, title=True)
    if buttons:
        for button in buttons:
            # 冲刺课程，跳过处理
            if '重点' in button.title:
                return False
        logger.info(f"Found commu event: {button.title}")
        logger.info("Select first choice")
        if buttons[0].selected:
            device.click(buttons[0])
        else:
            device.double_click(buttons[0])
        return True
    return False
    

if __name__ == '__main__':
    from logging import getLogger
    import logging
    logging.basicConfig(level=logging.INFO, format='[%(asctime)s] [%(levelname)s] [%(name)s] %(message)s')
    getLogger('kotonebot').setLevel(logging.DEBUG)
    getLogger(__name__).setLevel(logging.DEBUG)

    select_p_item()


import re
import time
import logging
import unicodedata
from functools import lru_cache
from dataclasses import dataclass
from typing_extensions import Self
from typing import Callable, NamedTuple

import cv2
import numpy as np
from cv2.typing import MatLike
from thefuzz import fuzz as _fuzz
from rapidocr_onnxruntime import RapidOCR


from .core import HintBox, Image, unify_image
from ..util import Rect, lf_path
from .debug import result as debug_result, debug

logger = logging.getLogger(__name__)

# TODO: 这个路径需要能够独立设置
_engine_jp = RapidOCR(
    rec_model_path=lf_path('models/japan_PP-OCRv4_rec_infer.onnx'),
    use_det=True,
    use_cls=False,
    use_rec=True,
)
_engine_en = RapidOCR(
    rec_model_path=lf_path('models/en_PP-OCRv3_rec_infer.onnx'),
    use_det=True,
    use_cls=False,
    use_rec=True,
)

StringMatchFunction = Callable[[str], bool]
REGEX_NUMBERS = re.compile(r'\d+')

@dataclass
class OcrResult:
    text: str
    rect: Rect
    confidence: float
    original_rect: Rect
    """
    识别结果在原图中的区域坐标。

    如果识别时没有设置 `rect` 或 `hint` 参数，则此属性值与 `rect` 相同。
    """

    def __repr__(self) -> str:
        return f'OcrResult(text="{self.text}", rect={self.rect}, confidence={self.confidence})'

    def replace(self, old: str, new: str, count: int = -1) -> Self:
        """
        替换识别结果中的文本。
        """
        self.text = self.text.replace(old, new, count)
        return self

    def regex(self, pattern: re.Pattern | str) -> list[str]:
        """
        提取识别结果中符合正则表达式的文本。
        """
        if isinstance(pattern, str):
            pattern = re.compile(pattern)
        return pattern.findall(self.text)

    def numbers(self) -> list[int]:
        """
        提取识别结果中的数字。
        """
        return [int(x) for x in REGEX_NUMBERS.findall(self.text)]

class OcrResultList(list[OcrResult]):
    def squash(self, remove_newlines: bool = True) -> OcrResult:
        """
        将所有识别结果合并为一个大结果。
        """
        if not self:
            return OcrResult('', (0, 0, 0, 0), 0, (0, 0, 0, 0))
        text = [r.text for r in self]
        confidence = sum(r.confidence for r in self) / len(self)
        points = []
        for r in self:
            points.append((r.rect[0], r.rect[1]))
            points.append((r.rect[0] + r.rect[2], r.rect[1]))
            points.append((r.rect[0], r.rect[1] + r.rect[3]))
            points.append((r.rect[0] + r.rect[2], r.rect[1] + r.rect[3]))
        rect = bounding_box(points)
        text = '\n'.join(text)
        if remove_newlines:
            text = text.replace('\n', '')
        return OcrResult(
            text=text,
            rect=rect,
            confidence=confidence,
            original_rect=rect,
        )

    def first(self) -> OcrResult | None:
        """
        返回第一个识别结果。
        """
        return self[0] if self else None

    def where(self, pattern: StringMatchFunction) -> 'OcrResultList':
        """
        返回符合条件的识别结果。
        """
        return OcrResultList([x for x in self if pattern(x.text)])

class TextNotFoundError(Exception):
    def __init__(self, pattern: str | re.Pattern | StringMatchFunction, image: 'MatLike'):
        self.pattern = pattern
        self.image = image
        super().__init__(f"Expected text not found: {pattern}")

class TextComparator:
    def __init__(self, name: str, text: str, func: Callable[[str], bool]):
        self.name = name
        self.text = text
        self.func = func

    def __call__(self, text: str) -> bool:
        return self.func(text)
    
    def __repr__(self) -> str:
        return f'{self.name}("{self.text}")'

@lru_cache(maxsize=1000)
def fuzz(text: str) -> TextComparator:
    """返回 fuzzy 算法的字符串匹配函数。"""
    func = lambda s: _fuzz.ratio(s, text) > 90
    return TextComparator("fuzzy", text, func)

@lru_cache(maxsize=1000)
def regex(regex: str) -> TextComparator:
    """返回正则表达式字符串匹配函数。"""
    func = lambda s: re.match(regex, s) is not None
    return TextComparator("regex", regex, func)

@lru_cache(maxsize=1000)
def contains(text: str, *, ignore_case: bool = False) -> TextComparator:
    """返回包含指定文本的函数。"""
    if ignore_case:
        func = lambda s: text.lower() in s.lower()
    else:
        func = lambda s: text in s
    return TextComparator("contains", text, func)

@lru_cache(maxsize=1000)
def equals(
    text: str,
    *,
    remove_space: bool = False,
    ignore_case: bool = True,
) -> TextComparator:
    """
    返回等于指定文本的函数。
    
    :param text: 要比较的文本。
    :param remove_space: 是否忽略空格。默认为 False。
    :param ignore_case: 是否忽略大小写。默认为 True。
    """
    def compare(s: str) -> bool:
        nonlocal text

        if ignore_case:
            text = text.lower()
            s = s.lower()
        if remove_space:
            text = text.replace(' ', '').replace('　', '')
            s = s.replace(' ', '').replace('　', '')

        return text == s
    return TextComparator("equals", text, compare)

def grayscaled(img: 'MatLike | str | Image') -> MatLike:
    img = unify_image(img)
    return cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

def _is_match(text: str, pattern: re.Pattern | str | StringMatchFunction | TextComparator) -> bool:
    if isinstance(pattern, re.Pattern):
        return pattern.match(text) is not None
    elif callable(pattern):
        return pattern(text)
    else:
        return text == pattern

# https://stackoverflow.com/questions/46335488/how-to-efficiently-find-the-bounding-box-of-a-collection-of-points
def _bounding_box(points):
    x_coordinates, y_coordinates = zip(*points)

    return [(min(x_coordinates), min(y_coordinates)), (max(x_coordinates), max(y_coordinates))]

def bounding_box(points: list[tuple[int, int]]) -> tuple[int, int, int, int]:
    """
    计算点集的外接矩形。

    :param points: 点集。以左上角为原点，向下向右为正方向。
    :return: 外接矩形的左上角坐标和宽高
    """
    topleft, bottomright = _bounding_box(points)
    return (topleft[0], topleft[1], bottomright[0] - topleft[0], bottomright[1] - topleft[1])

def pad_to(img: MatLike, target_size: int, rgb: tuple[int, int, int] = (255, 255, 255)) -> tuple[MatLike, tuple[int, int]]:
    """
    将图像居中填充到指定大小。缺少部分使用指定颜色填充。
    
    :return: 填充后的图像和填充的偏移量 (x, y)。
    """
    h, w = img.shape[:2]
    
    # 计算需要填充的宽高
    pad_h = max(0, target_size - h)
    pad_w = max(0, target_size - w)
    
    # 如果不需要填充则直接返回
    if pad_h == 0 and pad_w == 0:
        return img, (0, 0)
    
    # 创建目标画布并填充
    if len(img.shape) == 2:
        # 灰度图像
        ret = np.full((h + pad_h, w + pad_w), rgb[0], dtype=np.uint8)
    else:
        # RGB图像
        ret = np.full((h + pad_h, w + pad_w, 3), rgb, dtype=np.uint8)
    
    # 将原图像居中放置
    if len(img.shape) == 2:
        ret[
            pad_h // 2:pad_h // 2 + h,
            pad_w // 2:pad_w // 2 + w] = img
    else:
        ret[
            pad_h // 2:pad_h // 2 + h,
            pad_w // 2:pad_w // 2 + w, :] = img
    return ret, (pad_w // 2, pad_h // 2)

def _draw_result(image: 'MatLike', result: list[OcrResult]) -> 'MatLike':
    import numpy as np
    from PIL import Image, ImageDraw, ImageFont
    
    # 转换为PIL图像
    result_image = cv2.cvtColor(image.copy(), cv2.COLOR_BGR2RGB)
    pil_image = Image.fromarray(result_image)
    draw = ImageDraw.Draw(pil_image, 'RGBA')
    
    # 加载字体
    try:
        font = ImageFont.truetype(lf_path('res/fonts/SourceHanSansHW-Regular.otf'), 16)
    except:
        font = ImageFont.load_default()
    
    for r in result:
        # 画矩形框
        draw.rectangle(
            [r.rect[0], r.rect[1], r.rect[0] + r.rect[2], r.rect[1] + r.rect[3]], 
            outline=(255, 0, 0), 
            width=2
        )
        
        # 获取文本大小
        text = r.text + f" ({r.confidence:.2f})"  # 添加置信度显示
        text_bbox = draw.textbbox((0, 0), text, font=font)
        text_width = text_bbox[2] - text_bbox[0]
        text_height = text_bbox[3] - text_bbox[1]
        
        # 计算文本位置
        text_x = r.rect[0]
        text_y = r.rect[1] - text_height - 5 if r.rect[1] > text_height + 5 else r.rect[1] + r.rect[3] + 5
        
        # 添加padding
        padding = 4
        bg_rect = [
            text_x - padding,
            text_y - padding,
            text_x + text_width + padding,
            text_y + text_height + padding
        ]
        
        # 画半透明背景
        draw.rectangle(
            bg_rect,
            fill=(0, 0, 0, 128)
        )
        
        # 画文字
        draw.text(
            (text_x, text_y),
            text,
            font=font,
            fill=(255, 255, 255)
        )
    
    # 转回OpenCV格式
    result_image = cv2.cvtColor(np.array(pil_image), cv2.COLOR_RGB2BGR)
    return result_image

class Ocr:
    def __init__(self, engine: RapidOCR):
        self.__engine = engine

    # TODO: 考虑缓存 OCR 结果，避免重复调用。
    def ocr(
        self,
        img: 'MatLike',
        *,
        rect: Rect | None = None,
        pad: bool = True,
    ) -> OcrResultList:
        """
        OCR 一个 cv2 的图像。注意识别结果中的**全角字符会被转换为半角字符**。


        :param rect: 如果指定，则只识别指定矩形区域。
        :param pad:
            是否将过小的图像（尺寸 < 631x631）的图像填充到 631x631。
            默认为 True。

            对于 PaddleOCR 模型，图片尺寸太小会降低准确率。
            将图片周围填充放大，有助于提高准确率，降低耗时。
        :return: 所有识别结果
        """
        if rect is not None:
            x, y, w, h = rect
            img = img[y:y+h, x:x+w]
        original_img = img
        if pad:
            # TODO: 详细研究哪个尺寸最佳，以及背景颜色、图片位置是否对准确率与耗时有影响
            # https://blog.csdn.net/YY007H/article/details/124973777
            original_img = img.copy()
            img, pos_in_padded_img = pad_to(img, 631)
        else:
            pos_in_padded_img = (0, 0)
        img_content = img
        result, elapse = self.__engine(img_content)
        if result is None:
            return OcrResultList()
        ret = []
        for r in result:
            # HACK: 识别结果中包含奇怪的符号，暂时替换掉
            text = unicodedata.normalize('NFKC', r[1]).replace('ą', 'a')
            # r[0] = [左上, 右上, 右下, 左下]
            # 这里有个坑，返回的点不一定是矩形，只能保证是四边形
            # 所以这里需要计算出四个点的外接矩形
            result_rect = tuple(int(x) for x in bounding_box(r[0])) # type: ignore
            # result_rect (x, y, w, h)
            if rect is not None:
                original_rect = (
                    result_rect[0] + rect[0] - pos_in_padded_img[0],
                    result_rect[1] + rect[1] - pos_in_padded_img[1],
                    result_rect[2],
                    result_rect[3]
                )
            else:
                original_rect = result_rect
            if not len(original_rect) == 4:
                raise ValueError(f'Invalid original_rect: {original_rect}')
            if not len(result_rect) == 4:
                raise ValueError(f'Invalid result_rect: {result_rect}')
            confidence = float(r[2])
            ret.append(OcrResult(
                text=text,
                rect=result_rect,
                original_rect=original_rect,
                confidence=confidence
            ))
        ret = OcrResultList(ret)
        if debug.enabled:
            result_image = _draw_result(img, ret)
            elapse = elapse or [0, 0, 0]
            debug_result(
                'ocr',
                [result_image, original_img],
                f"pad={pad}\n" + \
                f"rect={rect}\n" + \
                f"elapsed: det={elapse[0]:.3f}s cls={elapse[1]:.3f}s rec={elapse[2]:.3f}s\n" + \
                f"result: \n" + \
                "<table class='result-table'><tr><th>Text</th><th>Confidence</th></tr>" + \
                "\n".join([f"<tr><td>{r.text}</td><td>{r.confidence:.3f}</td></tr>" for r in ret]) + \
                "</table>"
            )
        return ret

    def find(
        self,
        img: 'MatLike',
        text: str | re.Pattern | StringMatchFunction,
        *,
        hint: HintBox | None = None,
        rect: Rect | None = None,
        pad: bool = True,
    ) -> OcrResult | None:
        """
        识别图像中的文本，并寻找满足指定要求的文本。

        :param hint: 如果指定，则首先只识别 HintBox 范围内的文本，若未命中，再全局寻找。
        :param rect: 如果指定，则只识别指定矩形区域。此参数优先级低于 `hint`。
        :param pad: 见 `ocr` 的 `pad` 参数。
        :return: 找到的文本，如果未找到则返回 None
        """
        if hint is not None:
            if ret := self.find(img, text, rect=hint):
                logger.debug(f"find: {text} SUCCESS [hint={hint}]")
                return ret
            logger.debug(f"find: {text} FAILED [hint={hint}]")
        
        start_time = time.time()
        results = self.ocr(img, rect=rect, pad=pad)
        end_time = time.time()
        target = None
        for result in results:
            if _is_match(result.text, text):
                target = result
                break
        logger.debug(
            f"find: {text} {'SUCCESS' if target else 'FAILED'} " + \
            f"[elapsed={end_time - start_time:.3f}s] [rect={rect}]"
        )
        return target

    def find_all(
        self,
        img: 'MatLike',
        texts: list[str | re.Pattern | StringMatchFunction],
        *,
        hint: HintBox | None = None,
        rect: Rect | None = None,
        pad: bool = True,
    ) -> list[OcrResult | None]:
        """
        识别图像中的文本，并寻找多个满足指定要求的文本。

        :return:
            所有找到的文本，结果顺序与输入顺序相同。
            若某个文本未找到，则该位置为 None。
        """
        # HintBox 处理
        if hint is not None:
            result = self.find_all(img, texts, rect=hint, pad=pad)
            if all(result):
                return result

        ret: list[OcrResult | None] = []
        ocr_results = self.ocr(img, rect=rect, pad=pad)
        logger.debug(f"ocr_results: {ocr_results}")
        for text in texts:
            for result in ocr_results:
                if _is_match(result.text, text):
                    ret.append(result)
                    break
            else:
                ret.append(None)
        return ret
    
    def expect(
        self,
        img: 'MatLike',
        text: str | re.Pattern | StringMatchFunction,
        *,
        hint: HintBox | None = None,
        rect: Rect | None = None,
        pad: bool = True,
    ) -> OcrResult:
        """
        识别图像中的文本，并寻找满足指定要求的文本。如果未找到则抛出异常。

        :param hint: 如果指定，则首先只识别 HintBox 范围内的文本，若未命中，再全局寻找。
        :param rect: 如果指定，则只识别指定矩形区域。此参数优先级高于 `hint`。
        :param pad: 见 `ocr` 的 `pad` 参数。
        :return: 找到的文本
        """
        ret = self.find(img, text, hint=hint, rect=rect, pad=pad)
        if ret is None:
            raise TextNotFoundError(text, img)
        return ret



jp = Ocr(_engine_jp)
"""日语 OCR 引擎。"""
en = Ocr(_engine_en)
"""英语 OCR 引擎。"""



if __name__ == '__main__':
    pass



import logging
from typing import NamedTuple
from datetime import timedelta
from .. import R
from kotonebot import action, ocr, device, regex

logger = logging.getLogger(__name__)

class AP(NamedTuple):
    current: int
    total: int
    next_refresh: timedelta

    def __repr__(self):
        return f'AP({self.current}/{self.total} {self.next_refresh})'

@action('获取当前 AP')
def ap() -> AP | None:
    texts = ocr.ocr(rect=R.Daily.BoxHomeAP)
    logger.info(f'BoxHomeAP ocr result: {texts}')
    # 当前 AP 和总 AP
    ap = texts.where(regex(r'\d+/\d+')).first()
    if not ap:
        logger.warning('AP not found.')
        return None
    current, total = ap.numbers()
    # 下一次刷新时间
    next_refresh = texts.where(regex(r'\d+:\d+')).first()
    if not next_refresh:
        logger.warning('Next refresh time not found.')
        return None
    next_refresh = timedelta(minutes=next_refresh.numbers()[0], seconds=next_refresh.numbers()[1])
    return AP(current=current, total=total, next_refresh=next_refresh)

@action('获取当前宝石')
def jewel() -> int | None:
    jewel = ocr.find(regex(r'[\d,]+'), rect=R.Daily.BoxHomeJewel)
    logger.info(f'BoxHomeJewel find result: {jewel}')
    if not jewel:
        logger.warning('Jewel not found.')
        return None
    return int(jewel.text.replace(',', '').replace('+', ''))


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=logging.INFO, format='[%(asctime)s] [%(levelname)s] [%(name)s] [%(funcName)s] [%(lineno)d] %(message)s')
    logging.getLogger('kotonebot').setLevel(logging.DEBUG)
    logger.setLevel(logging.DEBUG)
    print(ap())
    print(jewel())



"""收取活动费"""
import logging

from . import R
from .common import conf
from .actions.scenes import at_home, goto_home
from kotonebot import task, device, image, color

logger = logging.getLogger(__name__)

@task('收取活动费', screenshot_mode='manual-inherit')
def acquire_activity_funds():
    if not conf().activity_funds.enabled:
        logger.info('Activity funds acquisition is disabled.')
        return

    if not at_home():
        goto_home()
    device.screenshot()
    if color.find('#ff1249', rect=R.Daily.BoxHomeActivelyFunds):
        logger.info('Claiming activity funds.')
        device.click(R.Daily.BoxHomeActivelyFunds)
        device.click(image.expect_wait(R.Common.ButtonClose))
        logger.info('Activity funds claimed.')
    else:
        logger.info('No activity funds to claim.')
    
    while not at_home():
        pass

if __name__ == '__main__':
    import logging
    logging.basicConfig(level=logging.INFO, format='[%(asctime)s] [%(levelname)s] [%(name)s] [%(funcName)s] [%(lineno)d] %(message)s')
    logger.setLevel(logging.DEBUG)
    acquire_activity_funds()



import logging
from itertools import cycle
from typing import Optional, Literal

from kotonebot.backend.context.context import wait
from kotonebot.ui import user
from kotonebot.util import Countdown, Interval
from kotonebot.backend.dispatch import SimpleDispatcher

from . import R
from .common import conf, PIdol
from .actions.scenes import at_home, goto_home
from .actions.in_purodyuusu import hajime_pro, hajime_regular, resume_regular_produce
from kotonebot import device, image, ocr, task, action, sleep, equals, contains

logger = logging.getLogger(__name__)

def format_time(seconds):
    minutes = int(seconds // 60)
    seconds = int(seconds % 60)
    return f"{minutes}m {seconds}s"

def unify(arr: list[int]):
    # 先对数组进行排序
    arr.sort()
    result = []
    i = 0
    while i < len(arr):
        # 将当前元素加入结果
        result.append(arr[i])
        # 跳过所有与当前元素相似的元素
        j = i + 1
        while j < len(arr) and abs(arr[j] - arr[i]) <= 10:
            j += 1
        i = j
    return result

@action('选择P偶像', screenshot_mode='manual-inherit')
def select_idol(target_titles: list[str] | PIdol):
    """
    选择目标P偶像

    前置条件：培育-偶像选择页面 1.アイドル選択\n
    结束状态：培育-偶像选择页面 1.アイドル選択\n

    :param target_titles: 目标偶像的名称关键字。选择时只会选择所有关键字都出现的偶像。
    """
    # 前置条件：[res/sprites/jp/produce/produce_preparation1.png]
    # 结束状态：[res/sprites/jp/produce/produce_preparation1.png]
    
    logger.info(f"Find and select idol: {target_titles}")
    # 进入总览
    device.screenshot()
    device.click(image.expect(R.Produce.ButtonPIdolOverview))
    while not image.find(R.Common.ButtonConfirmNoIcon):
        device.screenshot()

    if isinstance(target_titles, PIdol):
        target_titles = target_titles.to_title()
    _target_titles = [contains(t, ignore_case=True) for t in target_titles]
    device.screenshot()
    # 定位滑动基准
    results = image.find_all(R.Produce.IconPIdolLevel)
    results.sort(key=lambda r: tuple(r.position))
    ys = unify([r.position[1] for r in results])

    min_y = ys[0]
    max_y = ys[1]

    found = False
    max_tries = 5
    tries = 0
    # TODO: 加入 ScrollBar 类，判断滚动条进度
    # 找到目标偶像
    while not found:
        # 首先检查当前选中的是不是已经是目标
        if all(ocr.find_all(_target_titles, rect=R.Produce.KbIdolOverviewName)):
            found = True
            break
        # 如果不是，就挨个选中，判断名称
        for r in results:
            device.click(r)
            sleep(0.3)
            device.screenshot(force=True)
            if all(ocr.find_all(_target_titles, rect=R.Produce.KbIdolOverviewName)):
                found = True
                break
        if not found:
            tries += 1
            if tries > max_tries:
                break
            # 翻页
            device.swipe(x1=100, x2=100, y1=max_y, y2=min_y)
            sleep(2)
            device.screenshot()
            results = image.find_all(R.Produce.IconPIdolLevel)
            results.sort(key=lambda r: tuple(r.position))

    device.click(image.expect(R.Common.ButtonConfirmNoIcon))
    return found

@action('培育开始.编成翻页', screenshot_mode='manual-inherit')
def select_set(index: int):
    """
    选择指定编号的支援卡/回忆编成。

    前置条件：STEP 2/3 页面
    结束状态：STEP 2/3 页面

    :param index: 支援卡/回忆编成的编号，从 1 开始。
    """
    def _current():
        numbers = []
        while not numbers:
            device.screenshot()
            numbers = ocr.ocr(rect=R.Produce.BoxSetCountIndicator).squash().numbers()
            if not numbers:
                logger.warning('Failed to get current set number. Retrying...')
                sleep(0.2)
        return numbers[0]
    
    max_retries = 3
    retry_count = 0
    
    while retry_count < max_retries:
        current = _current()
        logger.info(f'Navigate to set #{index}. Now at set #{current}.')
        
        # 计算需要点击的次数
        click_count = abs(index - current)
        if click_count == 0:
            logger.info(f'Already at set #{current}.')
            return
        click_target = R.Produce.PointProduceNextSet if current < index else R.Produce.PointProducePrevSet
        
        # 点击
        for _ in range(click_count):
            device.click(click_target)
            sleep(0.1)
        
        # 确认
        final_current = _current()
        if final_current == index:
            logger.info(f'Arrived at set #{final_current}.')
            return
        else:
            retry_count += 1
            logger.warning(f'Failed to navigate to set #{index}. Current set is #{final_current}. Retrying... ({retry_count}/{max_retries})')
    
    logger.error(f'Failed to navigate to set #{index} after {max_retries} retries.')
    
@action('继续当前培育')
def resume_produce():
    """
    继续当前培育

    前置条件：游戏首页，且当前有进行中培育\n
    结束状态：游戏首页
    """
    # 点击 プロデュース中
    # [res/sprites/jp/daily/home_1.png]
    logger.info('Click ongoing produce button.')
    device.click(R.Produce.BoxProduceOngoing)
    # 点击 再開する
    # [res/sprites/jp/produce/produce_resume.png]
    logger.info('Click resume button.')
    device.click(image.expect_wait(R.Produce.ButtonResume))
    # 继续流程
    resume_regular_produce()

@action('执行培育', screenshot_mode='manual-inherit')
def do_produce(
    idol: PIdol,
    mode: Literal['regular', 'pro'],
    memory_set_index: Optional[int] = None
) -> bool:
    """
    进行培育流程

    前置条件：可导航至首页的任意页面\n
    结束状态：游戏首页\n
    
    :param idol: 要培育的偶像。如果为 None，则使用配置文件中的偶像。
    :param mode: 培育模式。
    :return: 是否因为 AP 不足而跳过本次培育。
    """
    if not at_home():
        goto_home()

    device.screenshot()
    # 有进行中培育的情况
    if ocr.find(contains('中'), rect=R.Produce.BoxProduceOngoing):
        logger.info('Ongoing produce found. Try to resume produce.')
        resume_produce()
        return True

    # 0. 进入培育页面
    mode_text = 'REGULAR' if mode == 'regular' else 'PRO'
    result = (SimpleDispatcher('enter_produce')
        .click(R.Produce.ButtonProduce)
        .click(contains(mode_text))
        .until(R.Produce.ButtonPIdolOverview, result=True)
        .until(R.Produce.TextAPInsufficient, result=False)
    ).run()
    if not result:
        logger.info('AP insufficient. Exiting produce.')
        device.click(image.expect_wait(R.InPurodyuusu.ButtonCancel))
        return False
    # 1. 选择 PIdol [screenshots/produce/select_p_idol.png]
    select_idol(idol.to_title())
    device.click(image.expect_wait(R.Common.ButtonNextNoIcon))
    # 2. 选择支援卡 自动编成 [screenshots/produce/select_support_card.png]
    ocr.expect_wait(contains('サポート'), rect=R.Produce.BoxStepIndicator)
    it = Interval()
    while True:
        if image.find(R.Common.ButtonNextNoIcon, colored=True):
            device.click()
            break
        elif image.find(R.Produce.ButtonAutoSet):
            device.click()
            sleep(1)
        elif image.find(R.Common.ButtonConfirm, colored=True):
            device.click()
        device.screenshot()
        it.wait()
    # 3. 选择回忆 自动编成 [screenshots/produce/select_memory.png]
    ocr.expect_wait(contains('メモリー'), rect=R.Produce.BoxStepIndicator)
    # 自动编成
    if memory_set_index is not None and not 1 <= memory_set_index <= 10:
        raise ValueError('`memory_set_index` must be in range [1, 10].')
    if memory_set_index is None:
        device.click(image.expect_wait(R.Produce.ButtonAutoSet))
        wait(0.5, before='screenshot')
        device.screenshot()
    # 指定编号
    else:
        select_set(memory_set_index)
    (SimpleDispatcher('do_produce.step_3')
        .click(R.Common.ButtonNextNoIcon)
        .click(R.Common.ButtonConfirm)
        .until(contains('開始確認'), rect=R.Produce.BoxStepIndicator)
    ).run()

    # 4. 选择道具 [screenshots/produce/select_end.png]
    # TODO: 如果道具不足，这里加入推送提醒
    if conf().produce.use_note_boost:
        if image.find(R.Produce.CheckboxIconNoteBoost):
            device.click()
            sleep(0.1)
    if conf().produce.use_pt_boost:
        if image.find(R.Produce.CheckboxIconSupportPtBoost):
            device.click()
            sleep(0.1)
    device.click(image.expect_wait(R.Produce.ButtonProduceStart))
    # 5. 相关设置弹窗 [screenshots/produce/skip_commu.png]
    cd = Countdown(5).start()
    while not cd.expired():
        device.screenshot()
        if image.find(R.Produce.RadioTextSkipCommu):
            device.click()
        if image.find(R.Common.ButtonConfirmNoIcon):
            device.click()
    if mode == 'regular':
        hajime_regular()
    else:
        hajime_pro()
    return True

@task('培育')
def produce_task(
    mode: Optional[Literal['regular', 'pro']] = None,
    count: Optional[int] = None,
    idols: Optional[list[PIdol]] = None,
    memory_sets: Optional[list[int]] = None
):
    """
    培育任务

    :param mode: 培育模式。若为 None，则从配置文件中读入。
    :param count: 培育次数。若为 None，则从配置文件中读入。
    :param idols: 要培育的偶像。若为 None，则从配置文件中读入。
    """
    if not conf().produce.enabled:
        logger.info('Produce is disabled.')
        return
    import time
    if count is None:
        count = conf().produce.produce_count
    if idols is None:
        idols = conf().produce.idols
    if memory_sets is None:
        memory_sets = conf().produce.memory_sets
    if mode is None:
        mode = conf().produce.mode
    # 数据验证
    if count < 0:
        user.warning('配置有误', '培育次数不能小于 0。将跳过本次培育。')
        return

    idol_iterator = cycle(idols)
    memory_set_iterator = cycle(memory_sets)
    for i in range(count):
        start_time = time.time()
        idol = next(idol_iterator)
        if conf().produce.auto_set_memory:
            memory_set = None
        else:
            memory_set = next(memory_set_iterator, None)
        logger.info(
            f'Produce start with: '
            f'idol: {idol.value}, mode: {mode}, memory_set: #{memory_set}'
        )
        if not do_produce(idol, mode, memory_set):
            user.info('AP 不足', f'由于 AP 不足，跳过了 {count - i} 次培育。')
            logger.info('%d produce(s) skipped because of insufficient AP.', count - i)
            break
        end_time = time.time()
        logger.info(f"Produce time used: {format_time(end_time - start_time)}")

if __name__ == '__main__':
    import logging
    logging.basicConfig(level=logging.INFO, format='[%(asctime)s] [%(levelname)s] [%(name)s] [%(funcName)s] [%(lineno)d] %(message)s')
    logging.getLogger('kotonebot').setLevel(logging.DEBUG)
    logger.setLevel(logging.DEBUG)
    import os
    from datetime import datetime
    os.makedirs('logs', exist_ok=True)
    log_filename = datetime.now().strftime('logs/task-%y-%m-%d-%H-%M-%S.log')
    file_handler = logging.FileHandler(log_filename, encoding='utf-8')
    file_handler.setFormatter(logging.Formatter('[%(asctime)s] [%(levelname)s] [%(name)s] %(message)s'))
    logging.getLogger().addHandler(file_handler)
    
    import time
    from kotonebot.backend.context import init_context, manual_context
    from kotonebot.tasks.common import BaseConfig
    from kotonebot.util import Profiler
    init_context(config_type=BaseConfig)
    conf().produce.enabled = True
    conf().produce.mode = 'pro'
    conf().produce.produce_count = 1
    conf().produce.idols = [PIdol.月村手毬_アイヴイ]
    conf().produce.memory_sets = [5]
    conf().produce.auto_set_memory = False
    # do_produce(PIdol.月村手毬_初声, 'pro', 5)
    produce_task()
    # a()
    # select_idol(PIdol.藤田ことね_学園生活)
    # select_set(10)
    # manual_context().begin()
    # print(ocr.ocr(rect=R.Produce.BoxSetCountIndicator).squash().numbers())


import requests
import os

def upload(file_path: str) -> str:
    """
    上传文件到 paste.sensio.no
    
    Args:
        file_path: 要上传的文件路径
    
    Returns:
        str: 上传后的 URL
    """
    url = 'https://paste.sensio.no/'
    headers = {
        'accept': 'text/plain',
        'User-Agent': 'KAA',
        'x-uuid': ''
    }
    
    files = {
        'file': (os.path.basename(file_path), open(file_path, 'rb'))
    }
        
    response = requests.post(url, files=files, headers=headers, allow_redirects=False)
    
    if response.status_code != 200:
        raise Exception(f"Upload failed with status code {response.status_code}")
    
    return response.text.strip()

if __name__ == "__main__":
    test_file = "version"
    if os.path.exists(test_file):
        result = upload(test_file)
        print(f"Upload result: {result}")



"""领取社团奖励，并尽可能地给其他人送礼物"""
import logging

from . import R
from .common import conf
from .actions.scenes import at_home, goto_home
from kotonebot.tasks.game_ui import toolbar_menu
from kotonebot import task, device, image, sleep, ocr

logger = logging.getLogger(__name__)

@task('领取社团奖励并送礼物')
def club_reward():
    """
    领取社团奖励，并尽可能地给其他人送礼物
    """

    if not conf().club_reward.enabled:
        logger.info('"Club reward" is disabled.')
        return
    
    if not at_home():
        goto_home()
    
    # 进入社团UI
    logger.info('Entering club UI')
    device.click(toolbar_menu(True))
    device.click(image.expect_wait(R.Daily.IconMenuClub, timeout=5))
    sleep(3)

    # 如果笔记请求尚未结束，则不进行任何笔记请求有关操作（领取奖励 & 发起新的笔记请求）

    # 如果笔记请求已经结束，且存在奖励提示，学偶UI应该会直接弹出面板，那么直接点击关闭按钮即可；
    logger.info('Prepare to collect note request reward')
    if image.find(R.Common.ButtonClose):
        device.click()
        logger.info('Collected note request reward')
    sleep(1)

    # 如果笔记请求已经结束，则发起一轮新的笔记请求；
    # 注：下面这个图片要可以区分出笔记请求是否已经结束，不然会发生不幸的事情
    logger.info('Prepare to start new note request')

    texts = ocr.ocr(rect=R.Daily.Club.NoteRequestHintBox)
    logger.debug(f'OCR result: {texts}')
    # 不应该进入的情况，识别结果为：[OcrResult(text="リクエストロ", rect=(243, 298, 145, 35), confidence=0.8576575517654419)]
    # 应该进入的情况，识别结果为：[OcrResult(text="リクエスト", rect=(244, 297, 141, 35), confidence=0.9993334531784057)]
    if texts and texts[0].text == 'リクエスト':
        # 经测验，threshold=0.999时也可以正确识别，所以这里保留这个阈值
        device.click(image.expect_wait(R.Daily.ButtonClubCollectReward, threshold=0.99))
        sleep(0.5)
        # 找到配置中选择的书籍
        device.click(image.expect_wait(conf().club_reward.selected_note.to_resource(), timeout=5))
        sleep(0.5)
        # 确认键
        device.click(image.expect_wait(R.Common.ButtonConfirm, timeout=5))
        sleep(0.5)
        device.click(image.expect_wait(R.Common.ButtonConfirm, timeout=5))
        sleep(1)
        logger.info('Started new note request')
    else:
        logger.info('No need to start new note request')
    
    # 送礼物（好友硬币是重要的o(*￣▽￣*)o
    logger.info('Sending gifts')
    for _ in range(5): # 默认循环5次
        # 送礼物
        if image.find(R.Daily.ButtonClubSendGift):
            device.click()
            sleep(0.5)
        # 下个人
        if image.find(R.Daily.ButtonClubSendGiftNext):
            device.click()
            sleep(0.5)
        else:
            # 找不到下个人就break
            break

if __name__ == '__main__':
    import logging
    logging.basicConfig(level=logging.INFO, format='[%(asctime)s] [%(levelname)s] [%(name)s] [%(funcName)s] [%(lineno)d] %(message)s')
    logger.setLevel(logging.DEBUG)
    club_reward()




class KotonebotError(Exception):
    pass

class KotonebotWarning(Warning):
    pass

class UnrecoverableError(KotonebotError):
    pass

class GameUpdateNeededError(UnrecoverableError):
    def __init__(self):
        super().__init__(
            'Game update required. '
            'Please go to Play Store and update the game manually.'
        )

class ResourceFileMissingError(KotonebotError):
    def __init__(self, file_path: str, description: str):
        self.file_path = file_path
        self.description = description
        super().__init__(f'Resource file ({description}) "{file_path}" is missing.')


import os
import json
import uuid
from typing import Any, Literal, TextIO

import cv2
from cv2.typing import MatLike

from kotonebot.util import cv2_imwrite

TraceId = Literal['rec-card']
TRACE_DIR = './traces/'
_trace_files: dict[TraceId, TextIO] = {}

if not os.path.exists(TRACE_DIR):
    os.makedirs(TRACE_DIR)

def trace(id: TraceId, image: MatLike, message: str | dict[str, Any]):
    file = None
    dir = os.path.join(TRACE_DIR, id)
    if id not in _trace_files:
        if not os.path.exists(dir):
            os.makedirs(dir)
        file = open(os.path.join(dir, id + '.log'), 'a+', encoding='utf-8')
        _trace_files[id] = file
    else:
        file = _trace_files[id]

    image_name = uuid.uuid4().hex
    cv2_imwrite(os.path.join(dir, image_name + '.png'), image)
    if isinstance(message, dict):
        message = json.dumps(message)
    message = f'{image_name}.png\n{message}\n'
    file.write(message)
    file.flush()



"""消息框、通知、推送等 UI 相关函数"""
import os
import time

import cv2
from cv2.typing import MatLike

from .pushkit import Wxpusher
from .. import logging

logger = logging.getLogger(__name__)

def retry(func):
    """
    装饰器：当函数发生 ConnectionResetError 时自动重试三次
    """
    def wrapper(*args, **kwargs):
        for i in range(3):
            try:
                return func(*args, **kwargs)
            except ConnectionResetError:
                if i == 2:  # 最后一次重试失败
                    raise
                logger.warning(f'ConnectionResetError raised when calling {func}, retrying {i+1}/{3}')
                continue
    return wrapper

def ask(
    question: str,
    options: list[str],
    *,
    timeout: float = -1,
) -> bool:
    """
    询问用户
    """
    raise NotImplementedError

def _save_local(
    title: str,
    message: str,
    images: list[MatLike] | None = None
):
    """
    保存消息到本地
    """
    if not os.path.exists('messages'):
        os.makedirs('messages')
    file_name = f'messages/{time.strftime("%Y-%m-%d_%H-%M-%S")}'
    with open(file_name + '.txt', 'w', encoding='utf-8') as f:
        logger.verbose('saving message to local: %s', file_name + '.txt')
        f.write(message)
    if images is not None:
        for i, image in enumerate(images):
            logger.verbose('saving image to local: %s', f'{file_name}_{i}.png')
            cv2.imwrite(f'{file_name}_{i}.png', image)

@retry
def push(
    title: str,
    message: str | None = None,
    *,
    images: list[MatLike] | None = None
):
    """
    推送消息
    """
    message = message or ''
    try:
        logger.verbose('pushing to wxpusher: %s', message)
        wxpusher = Wxpusher()
        wxpusher.push(title, message, images=images)
    except Exception as e:
        logger.warning('push remote message failed: %s', e)
        _save_local(title, message, images)

def info(
    title: str,
    message: str | None = None,
    images: list[MatLike] | None = None,
    *,
    once: bool = False
):
    logger.info('user.info: %s', message)
    push('KAA：' + title, message, images=images)

def warning(
    title: str,
    message: str | None = None,
    images: list[MatLike] | None = None,
    *,
    once: bool = False
):
    """
    警告信息。

    :param message: 消息内容
    :param once: 每次运行是否只显示一次。
    """
    logger.warning('user.warning: %s', message)
    push("KAA 警告：" + title, message, images=images)

def error(
    title: str,
    message: str | None = None,
    images: list[MatLike] | None = None,
    *,
    once: bool = False
):
    """
    错误信息。
    """
    logger.error('user.error: %s', message)
    push("KAA 错误：" + title, message, images=images)



"""升级一张支援卡，优先升级低等级支援卡"""
import logging

from kotonebot import task, device, image, sleep
from . import R
from .common import conf
from .actions.scenes import at_home, goto_home

logger = logging.getLogger(__name__)

@task('升级一张低等级支援卡')
def upgrade_support_card():
    """
    升级一张支援卡，优先升级低等级支援卡

    自动化思路是这样的：
    进入支援卡页面后，一直往下滑，滑倒底部（低等级支援卡区域）；
    然后点击左上角第一张支援卡，将左上角第一张支援卡提升一级。
    """

    if not conf().upgrade_support_card.enabled:
        logger.info('"Upgrade support card" is disabled.')
        return
    
    if not at_home():
        goto_home()
    
    # 进入支援卡页面
    logger.info('Entering Support Card page')
    device.click(image.expect_wait(R.Common.ButtonIdol, timeout=5))
    device.click(image.expect_wait(R.Common.ButtonIdolSupportCard, timeout=5))
    sleep(2)

    # TODO: 将下面硬编码的值塞到合适的地方

    # 往下滑，划到最底部
    for _ in range(5):
        device.swipe(
            R.Daily.SupportCard.DragDownStartPoint.x,
            R.Daily.SupportCard.DragDownStartPoint.y,
            R.Daily.SupportCard.DragDownEndPoint.x,
            R.Daily.SupportCard.DragDownEndPoint.y,
            duration=1.0
        )
        sleep(0.1)
    sleep(1.5)
    
    # 点击左上角第一张支援卡
    # 点击位置百分比: (0.18, 0.34)
    # 720p缩放后的位置: (130, 435)
    for _ in range(2):
        device.click(
            R.Daily.SupportCard.TargetSupportCard.x,
            R.Daily.SupportCard.TargetSupportCard.y
        )
        sleep(0.5)
    
    # 点击两次升级按钮（两个按钮的logo不一样，但是文字是一样的，这里资源文件只包含文字）
    device.click(image.expect_wait(R.Daily.ButtonSupportCardUpgrade, timeout=5))
    sleep(0.5)
    device.click(image.expect_wait(R.Daily.ButtonSupportCardUpgrade, timeout=5))
    sleep(1)

if __name__ == '__main__':
    import logging
    logging.basicConfig(level=logging.INFO, format='[%(asctime)s] [%(levelname)s] [%(name)s] [%(funcName)s] [%(lineno)d] %(message)s')
    logger.setLevel(logging.DEBUG)
    upgrade_support_card()




import logging
from typing import Callable



from .. import R
from .loading import loading
from kotonebot.util import Interval
from ..game_ui import toolbar_home
from kotonebot import device, image, action, cropped, until, sleep
from kotonebot.errors import UnrecoverableError

logger = logging.getLogger(__name__)


@action('检测是否位于首页')
def at_home() -> bool:
    return image.find(R.Daily.ButtonHomeCurrent) is not None

@action('检测是否位于日常商店页面')
def at_daily_shop() -> bool:
    icon = image.find(R.Daily.IconShopTitle)
    if icon is not None:
        return True
    else:
        # 调整默认购买数量的设置弹窗
        # [screenshots/contest/settings_popup.png]
        if image.find(R.Common.ButtonIconClose):
            device.click()
            sleep(1)
            return at_daily_shop()
        else:
            return False

@action('返回首页', screenshot_mode='manual-inherit')
def goto_home():
    """
    从其他场景返回首页。

    前置条件：无 \n
    结束状态：位于首页
    """
    logger.info("Going home.")
    it = Interval()
    while True:
        device.screenshot()
        if at_home():
            logger.info("At home.")
            break
        if image.find(R.Common.ButtonHome):
            device.click()
            logger.debug("Clicked home button.")
            sleep(0.2)
        elif home := toolbar_home():
            device.click(home)
            logger.debug("Clicked toolbar home button.")
            sleep(1)
        # 課題CLEAR [screenshots/go_home/quest_clear.png] 
        elif image.find(R.Common.ButtonIconClose):
            device.click()
            logger.debug("Clicked close button.")
            sleep(0.2)
        logger.debug(f"Trying to go home...")
        it.wait()

@action('前往商店页面')
def goto_shop():
    """
    从首页进入 ショップ。

    前置条件：无 \n
    结束状态：位于商店页面
    """
    logger.info("Going to shop.")
    if not at_home():
        goto_home()
    device.click(image.expect(R.Daily.ButtonShop))
    until(at_daily_shop, critical=True)

if __name__ == "__main__":
    import time
    goto_home()




"""检测与跳过交流"""
import logging

from cv2.typing import MatLike


from .. import R
from ..game_ui import WhiteFilter
from kotonebot.util import Countdown, Interval
from kotonebot import device, image, color, user, rect_expand, until, action, sleep, use_screenshot

logger = logging.getLogger(__name__)


@action('检查是否处于交流')
def is_at_commu():
    return image.find(R.Common.ButtonCommuSkip, preprocessors=[WhiteFilter()]) is not None

@action('跳过交流')
def skip_commu():
    device.click(image.expect_wait(R.Common.ButtonCommuSkip))

@action('检查未读交流', screenshot_mode='manual')
def handle_unread_commu(img: MatLike | None = None) -> bool:
    """
    检查当前是否处在未读交流，并自动跳过。

    :param img: 截图。
    :return: 是否跳过了交流。
    """
    ret = False
    logger.info('Check and skip commu')
    img = use_screenshot(img)
    skip_btn = image.find(R.Common.ButtonCommuFastforward, preprocessors=[WhiteFilter()])
    if skip_btn is None:
        logger.info('No fast forward button found. Not at a commu.')
        return ret
    
    ret = True
    logger.debug('Fast forward button found. Check commu')

    it = Interval()
    while True:
        if not is_at_commu():
            break
        if image.find(R.Common.ButtonCommuSkip, preprocessors=[WhiteFilter()]):
            device.click()
            logger.debug('Clicked skip button.')
        if image.find(R.Common.ButtonConfirm):
            logger.info('Unread commu found.')
            device.click()
            logger.debug('Clicked confirm button.')
            logger.debug('Pushing notification...')
            user.info('发现未读交流', images=[img])

        logger.debug('Fast forwarding...')
        it.wait()
    
    logger.info('Fast forward done.')
    return ret


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=logging.INFO, format='[%(asctime)s] [%(levelname)s] [%(name)s] [%(funcName)s] [%(lineno)d] %(message)s')
    logger.setLevel(logging.DEBUG)
    from kotonebot.backend.context import manual_context, inject_context
    from kotonebot.backend.debug.mock import MockDevice
    manual_context().begin()
    device = MockDevice()
    device.load_image(r"D:\current_screenshot.png")
    inject_context(device=device)
    print(is_at_commu())
    # while True:
    #     print(handle_unread_commu())
